from abc import ABC, abstractmethod
from typing import Tuple, Optional

from cyst.api.environment.message import Request, Response, Message
from cyst.api.host.service import Service
from cyst.api.network.node import Node


class PlatformInterface(ABC):
    """
    An interface that enables platforms to communicate with the core to elicit effects.
    """

    @abstractmethod
    def execute_task(self, task: Message, service: Optional[Service] = None, node: Optional[Node] = None, delay: float = 0.0) -> Tuple[bool, float]:
        """
        Instruct the core to execute included action with the appropriate behavioral model. The function signature
        follows the signature of :func:`cyst.api.logic.behavioral_model.action_effect`.

        :param task: The message containing the action. In virtually all cases this will be messages of type
            :class:`MessageType.REQUEST`, however, there is nothing preventing a platform to call behavioral models
            for actions on responses.
        :type task: Message

        :param service: A service on which the action should be executed. The platform is free to ignore this parameter
            if its associated behavioral model does not require it.
        :type service: Optional[Service]

        :param node: A node on which the action should be executed. The platform is free to ignore this parameter
            if its associated behavioral model does not require it.
        :type node: Optional[Node]

        :param delay: A delay before the execution of the action.
        :type delay: float

        :return: A tuple indicating whether the action was successfully executed and how long the execution took.
        """

    @abstractmethod
    def process_response(self, response: Response) -> Tuple[bool, int]:
        pass



from abc import ABC, abstractmethod
from typing import Callable, List, Tuple
from dataclasses import dataclass
from deprecated.sphinx import deprecated

from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.stores import ActionStore
from cyst.api.environment.message import Request, Response
from cyst.api.network.node import Node


@deprecated(version="0.6.0", reason="Replaced by behavioral models.")
class ActionInterpreter(ABC):
    """
    An action interpreter provides a semantic to the actions. It analyzes the incoming requests and then make
    appropriate changes in the environment and prepares responses that are sent back to the requests' origin.
    """
    @abstractmethod
    def evaluate(self, message: Request, node: Node) -> Tuple[int, Response]:
        """
        Evaluates the request.

        :param message: The request containing an action to be done.
        :type message: Request

        :param node: An instance of the node, where the message ended.
        :type node: Node

        :return: A tuple indicating the length of processing in the simulated time units and the Response that should
            be sent to the origin.
        """


@deprecated(version="0.6.0", reason="Replaced by behavioral models.")
@dataclass
class ActionInterpreterDescription:
    """ A definition of an action interpreter.

    :param namespace: A namespace in which all interpreted actions belong to. An arbitrary namespace nesting is
        supported through the dot notation (e.g., "a.b.c" namespaces and actions of the form "a.b.c.action_n").
    :type namespace: str

    :param description: A textual description of the action interpreter. The description should introduce the behavioral
        model that the interpreter implements, so that the users can decide whether to use it or not.
    :type description: str

    :param creation_fn: A factory function that can create the interpreter.
    :type creation_fn: Callable[[EnvironmentConfiguration, EnvironmentResources, EnvironmentPolicy, EnvironmentMessaging], ActionInterpreter]

    """
    namespace: str
    description: str
    creation_fn: Callable[[EnvironmentConfiguration, EnvironmentResources, EnvironmentPolicy, EnvironmentMessaging], ActionInterpreter]



from dataclasses import dataclass, field
from typing import List
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.network.firewall import FirewallRule, FirewallChainType, FirewallPolicy


@serialize(type_check=coerce)
@dataclass
class FirewallChainConfig(ConfigItem):
    """ Configuration of firewall chain.

    A firewall chain represents a set of rules that are applied to a network traffic.

    :param type: A type of traffic this chain applies to.
    :type type: FirewallChainType

    :param policy: A default policy applied to a traffic, which does not satisfy any rule.
    :type policy: FirewallPolicy

    :param rules: A set of rules governing what happens with a network traffic.
    :type rules: List[FirewallRule]
    """
    type: FirewallChainType
    policy: FirewallPolicy
    rules: List[FirewallRule]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__firewall_chain"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class FirewallConfig(ConfigItem):
    """ Configuration of a firewall.

    Firewall is represented as a collection of chains, with a default policy that is applied, unless specified
    otherwise.

    :param default_policy: A default policy applied to a traffic, which is not handled by a chain with its own policy.
    :type default_policy: FirewallPolicy

    :param chains: A list of firewall chain configurations.
    :type chains: List[FirewallChainConfig]
    """
    default_policy: FirewallPolicy
    chains: List[FirewallChainConfig]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__firewall"
    id: str = ""



from netaddr import IPNetwork, IPAddress

from cyst.api.environment.environment import Environment
from cyst.api.configuration import NodeConfig, PassiveServiceConfig, AccessLevel, ExploitConfig, VulnerableServiceConfig,\
                                   RouterConfig, InterfaceConfig, ConnectionConfig, ActiveServiceConfig
from cyst.api.host.service import Service
from cyst.api.network.node import Node
from cyst.api.logic.exploit import ExploitLocality, ExploitCategory

from cyst_services.scripted_actor.main import ScriptedActorControl

target = NodeConfig(
    active_services=[],
    passive_services=[
        PassiveServiceConfig(
            type="bash",
            owner="root",
            version="8.1.0",
            access_level=AccessLevel.LIMITED,
            local=True,
            id="bash_service"
        ),
        PassiveServiceConfig(
            type="lighttpd",
            owner="www",
            version="1.4.62",
            access_level=AccessLevel.LIMITED,
            local=False,
            id="web_server"
        )
    ],
    shell="bash",
    interfaces=[],
    traffic_processors=[],
    id="target"
)

attacker = NodeConfig(
    active_services=[
        ActiveServiceConfig(
            type="scripted_actor",
            name="attacker",
            owner="attacker",
            access_level=AccessLevel.LIMITED,
            id="attacker_service"
        )
    ],
    passive_services=[],
    interfaces=[],
    traffic_processors=[],
    shell="",
    id="attacker_node"
)

exploit1 = ExploitConfig(
    services=[
        VulnerableServiceConfig(
            name="lighttpd",
            min_version="1.4.62",
            max_version="1.4.62"
        )
    ],
    locality=ExploitLocality.REMOTE,
    category=ExploitCategory.CODE_EXECUTION,
    id="http_exploit"
)

router = RouterConfig(
    interfaces=[
      InterfaceConfig(
        ip=IPAddress("192.168.0.1"),
        net=IPNetwork("192.168.0.1/24"),
        index=0
      ),
      InterfaceConfig(
        ip=IPAddress("192.168.0.1"),
        net=IPNetwork("192.168.0.1/24"),
        index=1
      )
    ],
    traffic_processors=[],
    id="router"
)

connection1 = ConnectionConfig(
        src_id="target",
        src_port=-1,
        dst_id="router",
        dst_port=0
)

connection2 = ConnectionConfig(
        src_id="attacker_node",
        src_port=-1,
        dst_id="router",
        dst_port=1
)

e = Environment.create().configure(target, exploit1, router, connection1, attacker, connection2)

attacker_service = e.configuration.general.get_object_by_id("attacker_node.attacker", Service).active_service
attacker_control = e.configuration.service.get_service_interface(attacker_service, ScriptedActorControl)

e.control.init()

# ----------------------------------------------------------------------------------------------------------------------
# Addition to 05_create_adversary.py
# ----------------------------------------------------------------------------------------------------------------------
e.control.add_pause_on_response("attacker_node.attacker")

# Store the actions
actions = {}
for action in e.resources.action_store.get_prefixed("cyst"):
    actions[action.id] = action


# Display available actions
for action in actions.values():
    print(f"{action.id} ({action.description})")

# Scan the network for usable targets
action = actions["cyst:test:echo_success"]
for ip in IPNetwork("192.168.0.1/28").iter_hosts():
    attacker_control.execute_action(str(ip), "", action)
    e.control.run()
    print(f"{ip}: {attacker_control.get_last_response().status}")

# Look for exploitable services at the target
action = actions["cyst:host:get_remote_services"]
attacker_control.execute_action("192.168.0.2", "", action)
e.control.run()

services = attacker_control.get_last_response().content
print("Available services at 192.168.0.2: ", services)

useful_exploits = []

for service in services:
    service_name = service[0]
    service_version = service[1]
    potential_exploits = e.resources.exploit_store.get_exploit(service=service_name)
    for exp in potential_exploits:
        min_version = exp.services[service[0]].min_version
        max_version = exp.services[service[0]].max_version

        if min_version <= service_version <= max_version:
            useful_exploits.append((service[0], exp))

for exploit in useful_exploits:
    service_name = exploit[0]
    actual_exploit = exploit[1]
    print(f"Exploitable service: {service_name}, exploit category: {actual_exploit.category}, exploit locality: {actual_exploit.locality}")

# Use the exploit to get access to the target machine
action = actions["cyst:compound:session_after_exploit"]
action.set_exploit(useful_exploits[0][1])
attacker_control.execute_action("192.168.0.2", useful_exploits[0][0], action)
e.control.run()

# With the access, get information about the target
session = attacker_control.get_last_response().session
action = e.resources.action_store.get("meta:inspect:node")
attacker_control.execute_action("192.168.0.2", "", action, session=session)
e.control.run()

node: Node = attacker_control.get_last_response().content
print(f"Services at the target: {node.services.keys()}, interfaces at the target: {node.ips}")
# ----------------------------------------------------------------------------------------------------------------------

e.control.commit()

stats = e.infrastructure.statistics
print(f"Run id: {stats.run_id}\nStart time real: {stats.start_time_real}\n"
      f"End time real: {stats.end_time_real}\nDuration virtual: {stats.end_time_virtual}")



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional
from uuid import uuid4

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.utils.duration import Duration


@dataclass
class PhysicalAccessConfig(ConfigItem):
    """
    Physical access configuration specifies, which user can access given physical location in a given time. The access
    control works in a default deny mode, i.e., unless the user is cleared for access, the physical location is
    off-limits.

    :param identity: A unique identifier of a user. An asterisk '*' can be used to denote that this configuration
        applies to any user.
    :type identity: str

    :param time_from: A specification of a time of a day (the date part is ignored) from which the access is granted.
        If no time is provided, then 00:00:00 is assumed.
    :type time_from: datetime | None

    :param time_to: A specification of a time of a day (the date part is ignored) to which the access is granted.
        If no time is provided, then 23:59:59 is assumed.
    :type time_to: datetime | None
    """
    identity: str
    time_from: datetime | None
    time_to: datetime | None
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__physical_access"
    id: str = ""


@dataclass
class PhysicalLocationConfig(ConfigItem):
    """
    Physical location groups together physical assets, which will most often be Nodes and Routers. While Connections
    have their physical manifestations, they are not considered physical for this purpose.

    :param assets: A list of items that are physically present at a given location. This is represented as a list of
        their IDs in the configuration.
    :type assets: List[str]

    :param access: A list of physical access configurations.
    :type access: List[PhysicalAccessConfig]

    """
    assets: List[str]
    access: List[PhysicalAccessConfig]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__physical_location"
    id: str = ""


@dataclass
class PhysicalConnectionConfig(ConfigItem):
    """
    Physical connection represents traversing between two locations. This can be anything from going through a door
    frame to boarding an intercontinental flight. Physical connections do not have access control, i.e., any user can
    attempt to cross the divide between two locations, but if they are denied access at the destination, it is assumed
    that the user will have to return back effectively wasting twice the time without accomplishing nothing.

    Physical connections are bidirectional, so there is no real difference between the origin and destination.

    N.B.: Situation when a user can have access to an infrastructure via wireless networks when travelling between
    locations is currently not considered.

    :param origin: An id of the first location.
    :type origin: str

    :param destination: An id of the second location.
    :type destination: str

    :param travel_time: A time needed to traverse the physical connection.
    :type travel_time: Duration

    """
    origin: str
    destination: str
    travel_time: Duration
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__physical_connection"
    id: str = ""



import uuid

from collections.abc import Iterable
from typing import List, Tuple, Optional
from netaddr import IPAddress

from cyst.api.network.node import Node
from cyst.api.network.session import Session

from cyst.platform.network.elements import Hop, Endpoint, Resolver


# The session represents an existing chain of connections, which can be traversed without authorization by its owner
class SessionImpl(Session):
    def __init__(self, owner: str, parent: Optional[Session] = None, path: Optional[List[Hop]] = None,
                 src_service: str = "", dst_service: str = "", resolver: Optional[Resolver] = None,
                 id: Optional[str] = None) -> None:
        if not id:
            self._id = str(uuid.uuid4())
        else:
            self._id = id

        self._enabled = True

        # TODO Remove owners. They don't work and the are not needed
        if not owner:
            raise Exception("Cannot create a session without an owner")

        self._owner = owner
        self._parent: Optional[SessionImpl] = None
        if parent:
            self._parent = SessionImpl.cast_from(parent)

        # TODO: Session ownership is a dead-end concept and needs to be removed
        # if self._parent and parent.owner != self._owner:
        #     raise Exception("Cannot link sessions with different owners")

        if not path:
            raise Exception("Cannot create a session without a path")

        self._path: List[Hop] = path

        self._src_service = src_service
        self._dst_service = dst_service

        # Resolve all IP addresses if possible
        for hop in self._path:
            if not hop.src.ip and resolver:
                hop.src.ip = resolver.resolve_ip(hop.src.id, hop.src.port)

            if not hop.dst.ip and resolver:
                hop.dst.ip = resolver.resolve_ip(hop.dst.id, hop.dst.port)

        if self._parent and self._parent.endpoint == self._path[-1].dst:
            raise Exception("Cannot create a session sharing an endpoint with a parent")

    class ForwardIterator(Iterable):
        def __init__(self, session: 'SessionImpl') -> None:
            self._session = session
            self._path_index = 0
            self._parent_iterator: Optional['ForwardIterator'] = None #type: ignore #Seems like mypy struggles with nested classes
            if session.parent:
                self._parent_iterator = SessionImpl.cast_from(session.parent).forward_iterator
            else:
                self._parent_iterator = None
            self._parent_traversed = False

        def has_next(self) -> bool:
            if self._parent_iterator and self._parent_iterator.has_next():
                return True

            if self._path_index != len(self._session.path_id):
                return True

            return False

        def __iter__(self):
            return self

        def __next__(self) -> Hop:
            if self._parent_traversed or not self._session.parent:
                if self._path_index != len(self._session.path_id):
                    result = self._session.path_id[self._path_index]
                    self._path_index += 1
                    return result
                else:
                    raise StopIteration
            else:
                if self._parent_iterator.has_next():
                    return self._parent_iterator.__next__()
                else:
                    self._parent_traversed = True
                    return self.__next__()

    class ReverseIterator(Iterable):
        def __init__(self, session: 'SessionImpl') -> None:
            self._session = session
            self._path_index = len(self._session.path_id) - 1
            self._parent_iterator: Optional['ForwardIterator'] = None #type: ignore #Seems like mypy struggles with nested classes
            if session.parent:
                self._parent_iterator = SessionImpl.cast_from(session.parent).reverse_iterator
            else:
                self._parent_iterator = None
            self._parent_traversing = False

        def has_next(self) -> bool:
            if self._path_index >= 0:
                return True

            elif self._parent_iterator:
                return self._parent_iterator.has_next()

            return False

        def __iter__(self):
            return self

        def __next__(self) -> Hop:
            if self._path_index >= 0:
                result = self._session.path_id[self._path_index]
                self._path_index -= 1
                return result.swap()
            else:
                if not self._parent_traversing:
                    self._parent_traversing = True
                    return self.__next__()
                else:
                    if self._parent_iterator.has_next():
                        result = self._parent_iterator.__next__()
                        return result
                    else:
                        raise StopIteration

    # ------------------------------------------------------------------------------------------------------------------
    # Session interface
    @property
    def owner(self) -> str:
        return self._owner

    @property
    def id(self) -> str:
        return str(self._id)

    @property
    def parent(self) -> Session:
        return self._parent

    @property
    def path(self) -> List[Tuple[Optional[IPAddress], Optional[IPAddress]]]:
        return [(x.src.ip, x.dst.ip) for x in self._path]

    @property
    def end(self) -> Tuple[IPAddress, str]:
        return self._path[-1].dst.ip, self._dst_service

    @property
    def start(self) -> Tuple[IPAddress, str]:
        if self._parent:
            return self._parent.start
        else:
            return self._path[0].src.ip, self._src_service

    @property
    def enabled(self) -> bool:
        return self._enabled

    def disable(self) -> None:
        self._enabled = False

    def enable(self) -> None:
        self._enabled = True

    def terminates_at(self, node: Node) -> bool:
        end_ip = self.end
        for iface in node.interfaces:
            if iface.ip == end_ip:
                return True
        return False

    # ------------------------------------------------------------------------------------------------------------------

    @property
    def forward_iterator(self) -> ForwardIterator:
        return SessionImpl.ForwardIterator(self)

    @property
    def reverse_iterator(self) -> ReverseIterator:
        return SessionImpl.ReverseIterator(self)

    # Endpoint is a destination node of the last path hop
    @property
    def endpoint(self) -> Endpoint:
        return self._path[-1].dst

    @property
    def startpoint(self) -> Endpoint:
        if self._parent:
            return self._parent.startpoint
        else:
            return self._path[0].src

    @property
    def path_id(self) -> List[Hop]:
        return self._path

    def __str__(self) -> str:
        full_path = self.path
        parent = self._parent
        while parent:
            full_path = parent.path + full_path
            parent = parent.parent

        path_repr = [str(self.start)]
        path_repr.extend([str(x[1]) for x in full_path])
        return "[ID: {}, Owner: {}, Path: ({}|{}|{})]".format(self.id, self.owner, self._src_service, path_repr, self._dst_service)

    def __repr__(self) -> str:
        full_path = self.path_id
        parent = self._parent
        while parent:
            full_path = parent.path_id + full_path
            parent = parent.parent

        path_repr = [str(self.start)]
        path_repr.extend([str(x[1]) for x in full_path])
        return "[ID: {}, Owner: {}, Path:({}|{}|{})]".format(self.id, self.owner, self._src_service, path_repr, self._dst_service)

    def __eq__(self, other: 'SessionImpl') -> bool:
        # Comparing to None, everything is False
        if not other:
            return False

        # Hard type check
        if not isinstance(other, SessionImpl):
            return False #MYPY: Do we need this, if we follow anotations?

        # Owner comparison - to be removed
        if self.owner != other.owner:
            return False

        # If one session has parent and the other does not they are not the same
        # Also, looking for a better way to write it
        if (self.parent is None and other.parent is not None) or \
           (self.parent is not None and other.parent is None):
            return False

        # If their parents don't match, they are not the same
        if self.parent and self.parent != other.parent:
            return False

        # If their paths don't match, they are not the same
        if self._path != other._path:
            return False

        # If their source or destination services don't match, they are not the same
        if self._src_service != other._src_service or self._dst_service != other._dst_service:
            return False

        return True

    @staticmethod
    def cast_from(o: Session) -> 'SessionImpl':
        if isinstance(o, SessionImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Session interface")



from dataclasses import dataclass

from cyst.api.environment.stats import Statistics


@dataclass
class StatisticsImpl(Statistics):
    run_id: str = ""
    configuration_id: str = ""
    start_time_real: float = 0.0
    end_time_real: float = 0.0
    end_time_virtual: int = 0

    @staticmethod
    def cast_from(o: Statistics) -> 'StatisticsImpl':
        if isinstance(o, StatisticsImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Statistics interface")



from typing import Type, Any, List, Optional

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.environment.configuration import GeneralConfiguration, ObjectType, ConfigurationObjectType
from cyst.api.environment.platform import Platform

from cyst.platform.environment.configurator import Configurator

# ----------------------------------------------------------------------------------------------------------------------
class GeneralConfigurationImpl(GeneralConfiguration):

    def __init__(self, platform: Platform, env_general_configuration: GeneralConfiguration) -> None:
        self._configurator = Configurator(platform)
        self._env_general_configuration = env_general_configuration

    # ------------------------------------------------------------------------------------------------------------------
    # Proxy methods back to the environment
    def get_configuration(self) -> List[ConfigItem]:
        return self._env_general_configuration.get_configuration()

    def save_configuration(self, indent: Optional[int]) -> str:
        return self._env_general_configuration.save_configuration(indent)

    def load_configuration(self, config: str) -> List[ConfigItem]:
        return self._env_general_configuration.load_configuration(config)

    def get_configuration_by_id(self, id: str, configuration_type: Type[ConfigurationObjectType]) -> ConfigurationObjectType:
        return self._env_general_configuration.get_configuration_by_id(id, configuration_type)

    # ------------------------------------------------------------------------------------------------------------------
    # Local methods
    def get_object_by_id(self, id: str, object_type: Type[ObjectType]) -> ObjectType:
        return self._configurator.get_object_by_id(id, object_type)

    def add_object(self, id: str, obj: Any) -> None:
        self._configurator.add_object(id, obj)

    # Not fancying type removal, but it is unimportant here and only pollutes the code
    def configure(self, *configs) -> None:
        self._configurator.configure(*configs)

    @staticmethod
    def cast_from(o: GeneralConfiguration) -> 'GeneralConfigurationImpl':
        if isinstance(o, GeneralConfigurationImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the GeneralConfiguration interface")



from abc import ABC, abstractmethod
from dataclasses import dataclass
from deprecated.sphinx import versionadded
from enum import Enum, auto
from flags import Flags
from netaddr import IPAddress
from typing import Any, Optional, Union, TypeVar, Type, Dict

from cyst.api.network.session import Session
from cyst.api.logic.access import Authorization, AuthenticationToken, AuthenticationTarget
from cyst.api.logic.action import Action
from cyst.api.logic.metadata import Metadata


class MessageType(Enum):
    """ The type of the message.

    Possible values:

        :TIMEOUT: The message is an information about expiration of a timeout.
        :REQUEST: The message is a request.
        :RESPONSE: The message is a response.
        :RESOURCE: The message carries a resource.
        :SIGNAL: The message is a signaling between the environment and components

    """
    TIMEOUT = 0
    REQUEST = 1
    RESPONSE = 2
    RESOURCE = 3
    SIGNAL = 4


class StatusOrigin(Enum):
    """ Indicates the origin of the status of the response.

    Possible values:

        :NETWORK: The original request was processed and evaluated at the network or router level.
        :NODE: The original request was processed and evaluated at the node level, without reaching a specific service.
        :SERVICE: The original request was processed and evaluated at the service level.
        :RESOURCE: The original request was for the resources.
        :SYSTEM: The original request couldn't be evaluated in the environment context. This indicates a system error.

    """
    NETWORK = 0
    NODE = 1
    SERVICE = 2
    RESOURCE = 3
    SYSTEM = 99


class StatusValue(Enum):
    """ Indicates the general result of the response.

    Possible values:

        :SUCCESS: The action intended in the original request was successfully carried out.
        :FAILURE: The action intended in the original request was not successful due to wrong parameter combination or
            values.
        :ERROR: The action intended in the original request was not successful because of missing parameters or because
            the action could not be evaluated.
        :PARTIAL: The action intended in the original request was partially successful and another response should be
            expected to arrive soon.

    """
    SUCCESS = 0
    FAILURE = 1
    ERROR = 2
    PARTIAL = 3


class StatusDetail(Flags):
    """ Status detail provides another introspection mechanism to active services into the nature of failures and
    errors. Status detail follows unified naming convention WHAT_WHY, where WHY is one of the following:

    * NOT_PROVIDED: WHAT was not passed as a parameter, even though it is required

    * NOT_EXISTING: WHAT does not exist within the context of current simulation run (e.g., service name, user name,
      etc.)

    * NOT_APPLICABLE: WHAT cannot be used (e.g., wrong authorization, wrong exploit parameters, etc.)

    * NOT_SUPPORTED: WHAT exists as a valid concept, but the target does not support it (e.g., attempting to open a
      session to a service that does not support it)

    * NEXT: WHAT was a correct step towards success, but another WHAT is required

    General:
        :UNKNOWN: There is no additional detail to be provided.

    NODE.FAILURE:
        :PRIVILEGES_NOT_APPLICABLE: A wrong authentication/authorization used.

    NODE.ERROR:
        :SERVICE_NOT_PROVIDED: No service was specified.
        :SERVICE_NOT_EXISTING: A service does not exist at the node.
        :SESSION_NOT_PROVIDED: A session was required but not provided.
        :SESSION_NOT_APPLICABLE: A wrong session was provided.

    SERVICE.FAILURE:
        :SESSION_CREATION_NOT_SUPPORTED: A session cannot be created with this service.
        :EXPLOIT_NOT_PROVIDED: An action required an exploit, but it was not provided.
        :EXPLOIT_NOT_APPLICABLE: An exploit could not be used in combination with an action to affect the service.
        :EXPLOIT_CATEGORY_NOT_APPLICABLE: Wrong exploit category was used (e.g., data manipulation exploit for privilege
            escalation)
        :EXPLOIT_LOCALITY_NOT_APPLICABLE: Wrong exploit locality was used (e.g., local exploit for remote action)
        :EXPLOIT_PARAMETER_NOT_PROVIDED: A required exploit parameter was missing.
        :EXPLOIT_PARAMETER_NOT_APPLICABLE: A provided parameter in exploit could not be used to affect the service.
        :AUTHORIZATION_NOT_PROVIDED: An authorization was missing when trying to access the service.
        :AUTHORIZATION_NOT_APPLICABLE: A wrong authorization was used to access the service.
        :AUTHENTICATION_NOT_PROVIDED: An authentication token was missing when trying to authenticate against the
            service.
        :AUTHENTICATION_NOT_APPLICABLE: A wrong authentication token was used when trying to authenticate against the
            service.
        :AUTHENTICATION_NEXT: A previous step in multi-factor authentication step was successful, another step must
            be undertaken.

    SYSTEM.ERROR:
        :ACTION_NOT_EXISTING: There is no interpreter able to interpret the given action.

    """
    UNKNOWN = ()
    # NODE.FAILURE
    PRIVILEGES_NOT_APPLICABLE = ()

    # NODE.ERROR
    SERVICE_NOT_PROVIDED = ()
    SERVICE_NOT_EXISTING = ()
    SESSION_NOT_PROVIDED = ()
    SESSION_NOT_APPLICABLE = ()

    # SERVICE.FAILURE
    SESSION_CREATION_NOT_SUPPORTED = ()
    EXPLOIT_NOT_PROVIDED = ()
    EXPLOIT_NOT_APPLICABLE = ()
    EXPLOIT_CATEGORY_NOT_APPLICABLE = ()
    EXPLOIT_LOCALITY_NOT_APPLICABLE = ()
    EXPLOIT_PARAMETER_NOT_PROVIDED = ()
    EXPLOIT_PARAMETER_NOT_APPLICABLE = ()
    AUTHORIZATION_NOT_PROVIDED = ()
    AUTHORIZATION_NOT_APPLICABLE = ()
    AUTHENTICATION_NOT_PROVIDED = ()
    AUTHENTICATION_NOT_APPLICABLE = ()
    AUTHENTICATION_NEXT = ()
    ACTION_PARAMETER_NOT_PROVIDED = ()
    ACTION_PARAMETER_NOT_APPLICABLE = ()

    # SERVICE.ERROR

    # SYSTEM.FAILURE

    # SYSTEM.ERROR
    ACTION_NOT_EXISTING = ()


@dataclass
class Status:
    """ Status is bound to responses, as it informs about the result of request processing.

    :param origin: The origin of the status.
    :type origin: StatusOrigin

    :param value: The value of the status.
    :type value: StatusValue

    :param detail: Additional specification of the status value.
    :type detail: StatusDetail
    """
    origin: StatusOrigin
    value: StatusValue
    detail: StatusDetail = StatusDetail.UNKNOWN

    def __str__(self) -> str:
        if self.detail != StatusDetail.UNKNOWN:
            result = "({}, {}, {})".format(self.origin.name, self.value.name, self.detail.name)
        else:
            result = "({}, {})".format(self.origin.name, self.value.name)
        return result


T = TypeVar('T', bound=Union['Request', 'Response', 'Resource', 'Timeout'])


class Message(ABC):
    """
    Message provides a mean to exchange information between elements of the simulation.
    """

    @property
    @abstractmethod
    def id(self) -> int:
        """
        A numerical identifier of the message, which is unique across a simulation run. The request and response share
        the same id to enable correct pairing.

        :rtype: int
        """

    @property
    @abstractmethod
    def type(self) -> MessageType:
        """
        A type of the message.

        :rtype: MessageType
        """

    @property
    @abstractmethod
    def src_ip(self) -> Optional[IPAddress]:
        """
        An IP address from which the message was sent. Messages not yet sent and timeouts do not have a source IP.

        :rtype: Optional[IPAddress]
        """

    @property
    @abstractmethod
    def dst_ip(self) -> Optional[IPAddress]:
        """
        An IP address where the message was sent to. Timeouts do not have a destination IP.

        :rtype: Optional[IPAddress]
        """

    @property
    @abstractmethod
    def src_service(self) -> Optional[str]:
        """
        A service that produced the message. In case of service-less origins, such as router processing, the service
        is empty.

        :rtype: Optional[str]
        """

    @property
    @abstractmethod
    def dst_service(self) -> str:
        """
        A service that is the ultimate target for the message.

        :rtype: str
        """

    @property
    @abstractmethod
    def session(self) -> Optional[Session]:
        """
        A session associated with the message.

        :rtype: Session
        """

    @property
    @abstractmethod
    def auth(self) -> Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]]:
        """
        An authentication/authorization associated with the message. In case of requests, it can either contain
        Authorizations or Authentication tokens provided by the origin. In case of responses, it can be the original
        Authorization provided by the origin, Authorization provided by the destination service as a result of an
        actions, or Authentication target if the request was a part of multi-factor authentication scheme message
        exchange.

        :rtype: Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]]
        """

    @property
    @abstractmethod
    def ttl(self) -> int:
        """
        Time-to-live of the message. After each pass through a router, this value gets decreased and the message gets
        discarded if it ever reaches 0. This mechanism exists to prevent endless resending of a message in case of bad
        network topology.

        :rtype: int
        """

    @property
    @abstractmethod
    def metadata(self) -> Metadata:
        """
        Metadata associated with the message. The metadata is supplied to the message by metadata providers (see
        :class:`cyst.api.environment.metadata_provider.MetadataProvider`). The metadata mechanism is intended to
        provide at least some information for active services, who are not able to see the action portion of a
        message (at least that is the plan, currently they can).

        :rtype: Metadata
        """

    @abstractmethod
    def set_metadata(self, metadata: Metadata) -> None:
        """
        Sets a metadata to the message. This call overwrites the original value, so, if there is a multitude of
        metadata providers, it is advisable to use :func:`add_metadata()`.

        :param metadata: A new value of the metadata.
        :type metdata: Metadata
        """

    @versionadded(version="0.6.0")
    @property
    @abstractmethod
    def platform_specific(self) -> Dict[str, Any]:
        """
        Provides access to platform-specific parameters of a message. Their contents can be arbitrary and are intended
        for assisting with message handling in a non CYST-simulated environments.

        :rtype: Dict[str, Any]
        """

    @abstractmethod
    def cast_to(self, type: Type[T]) -> T:
        """
        This function enables typecasting of Message into one of the four derived types: Request, Response, Resource,
        and Timeout. While technically not necessary due to Python's type system, it conveys intention and also makes
        a check whether the conversion can actually be done.

        :param type: A type to cast the mesasge to.
        :type type: Type[TypeVar('T', bound=Union['Request', 'Response', 'Resource', 'Timeout'])]

        :rtype: T
        """


class Request(Message, ABC):
    """
    Request is a message specialization that carries an action.
    """

    @property
    @abstractmethod
    def action(self) -> Action:
        """
        Gets an action associated with the request.

        :rtype: Action
        """


class Response(Message, ABC):
    """
    Response is a message specialization the carries the result of associated request.
    """

    @versionadded(version="0.6.0")
    @property
    @abstractmethod
    def action(self) -> Action:
        """
        Gets an action associated with the original request.

        :rtype: Action
        """
        pass

    @property
    @abstractmethod
    def status(self) -> Status:
        """
        A result of processing the associated request.

        :rtype: Status
        """

    @property
    @abstractmethod
    def content(self) -> Optional[Any]:
        """
        Any associated information that was included during the processing of the associated request.

        :rtype: Optional[Any]
        """


class Timeout(Message, ABC):
    """
    Timeout is a message specialization that is delivered when a timeout requested through the
    :class:`cyst.api.environment.clock.Clock` interface.
    """
    @property
    @abstractmethod
    def start_time(self) -> float:
        """
        A simulation time when the timeout was requested.

        :rtype: int
        """

    @property
    @abstractmethod
    def duration(self) -> float:
        """
        A duration of the timeout in simulated time units.

        :rtype: int
        """

    @property
    @abstractmethod
    def parameter(self) -> Any:
        """
        Any parameter that was included during the invocation of the timeout function.

        :rtype: Any
        """


class Resource(Message, ABC):
    """
    Resource is a message specialization that is delivered when an asynchronous external resource was requested.
    """
    @property
    @abstractmethod
    def path(self) -> str:
        """
        A URL of the resource.

        :rtype str:
        """

    @property
    @abstractmethod
    def status(self) -> Status:
        """
        A result of processing the associated request.

        :rtype: Status
        """


    @property
    @abstractmethod
    def data(self) -> Optional[str]:
        """
        Data retrieved from the external resource. In case the system failed to extract the data (due to timeout or
        error), the return is set to None.

        :rtype: Optional[str]
        """


class ComponentState(Flags):
    """
    Component state represents an abstract state of a particular component within simulation. The state is expressed
    as a flag, so, multiple states can be expressed at once.

    States related to the component lifetime:
        :CREATED: A component was just created.
        :INIT: A component was successfully initialized.
        :RUNNING: A component is running in a nominal fashion.
        :PAUSED: An execution of component's functions was paused.
        :STOPPED: An execution of component's functions was stopped.
        :FINISHED: A component stopped its operation in a normal way.
        :TERMINATED: A component stopped its operation through some external factor.

    States related to action effects:
        :UNDER_ATTACK: A component is a subject to hostile activity. Note that there is no state indicating that attack has ended.
        :UNDER_CONTROL: A component was subject to hostile activity that resulted in access or control from the origin of activity.
        :RESTORED: A component is no longer UNDER_CONTROL.

    States related to goal pursuit:
        :GOAL_REACHED: A component reached its stated goal.
        :GOAL_UNREACHABLE: A component is unable to reach its goal.

    An example of a state evolution of a service under a successful attack:

        ComponentState.UNDER_ATTACK
        ComponentState.UNDER_CONTROL
        ComponentState.TERMINATED

    An example of a state evolution of a service that was attacked and cleaned afterward:

        ComponentState.UNDER_ATTACK
        ComponentState.UNDER_CONTROL
        ComponentState.STOPPED
        ComponentState.RUNNING | ComponentState.RESTORED

    An attacker signalling that it reached its goal and do not wish to continue the run:

        ComponentState.FINISHED | ComponentState.GOAL_REACHED

    """
    CREATED = ()
    INIT = ()
    RUNNING = ()
    PAUSED = ()
    STOPPED = ()
    FINISHED = ()
    TERMINATED = ()
    # Action-related state
    UNDER_ATTACK = ()
    UNDER_CONTROL = ()
    RESTORED = ()
    # Goal-related state
    GOAL_REACHED =()
    GOAL_UNREACHABLE = ()


class Signal(Message, ABC):
    """
    Signal is a message specialization that is used for communicating state changes between the environment and
    simulation components.
    """
    @property
    @abstractmethod
    def signal_origin(self) -> str:
        """
        An identification of a source of the signal. This will usually be the ID of a component, or "__environment" if
        the signal is coming from within CYST.

        :rtype: str
        """

    @property
    @abstractmethod
    def state(self) -> ComponentState:
        """
        A new state that the component entered.

        :rtype: ComponentState
        """

    @property
    @abstractmethod
    def effect_origin(self) -> str:
        """
        An identification of a source of the state change that prompted the signal emission.

        :rtype: str
        """

    @property
    @abstractmethod
    def effect_message(self) -> Optional[int]:
        """
        An ID of a message that caused the state change. In case of environment signals, there does not have to be any
        related message (such as the environment termination).

        :rtype: Optional[int]
        """

    @property
    @abstractmethod
    def effect_description(self) -> str:
        """
        A description of an effect used mainly for later analysis. For example "Maximum number of actions reached",
        or "Successful execution of an action x:y". This is not expected to be machine-processable .

        :rtype: str
        """

    @property
    @abstractmethod
    def effect_parameters(self) -> Dict[str, Any]:
        """
        Additional information related to the state change. For example, an agent may use this to signal its perceived
        reward, or some important training/execution parameters.

        :rtype: Dict[str, Any]
        """



import dataclasses
from dataclasses import dataclass, field
from typing import List, Union, Optional
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig
from cyst.api.configuration.network.elements import InterfaceConfig
from cyst.api.configuration.network.firewall import FirewallConfig


@serialize(type_check=coerce)
@dataclass
class NodeConfig(ConfigItem):
    """ Configuration for a network node

    A network node is a main building block of simulation topology. It is modelled as a collection of services and a
    set of network interfaces.

    :param active_services: A list of active services running on the node.
    :type active_services: List[Union[ActiveServiceConfig, str]]

    :param passive_services: A list of passive services running on the node.
    :type passive_services: List[Union[PassiveServiceConfig, str]]

    :param traffic_processors: A list of either active services that are acting as traffic processors, or at most one
        firewall.
    :type traffic_processors: Optional[List[Union[FirewallConfig, ActiveServiceConfig, str]]]

    :param shell: A name of a passive service that is designated as a shell. Properties of the shell service determine
        the impact of some exploits, e.g., when related to access rights. Technically, any name could be selected as a
        shell, because the system currently does not check whether the service exists or not.
    :type shell: str

    :param interfaces: A list of interfaces on the node.
    :type interfaces: List[Union[InterfaceConfig, str]]
    """
    active_services: List[Union[ActiveServiceConfig, str]]
    passive_services: List[Union[PassiveServiceConfig, str]]
    traffic_processors: List[Union[FirewallConfig, ActiveServiceConfig, str]]
    shell: str
    interfaces: List[Union[InterfaceConfig, str]]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__node"
    id: str = ""



import networkx as nx

from abc import ABC, abstractmethod
from netaddr import IPAddress
from typing import Optional, List, Dict, Union
from itertools import product

from cyst.api.environment.configuration import GeneralConfiguration
from cyst.api.network.node import Node

from cyst.platform.network.elements import Connection, ConnectionImpl, Hop, Endpoint, Resolver, InterfaceImpl
from cyst.platform.network.router import Router
from cyst.platform.network.node import NodeImpl
from cyst.platform.network.session import Session, SessionImpl


# TODO: The network is largely useless after moving the object store into configuration. I should probably think about
#       whether to keep it or not
class Network(Resolver):
    def __init__(self, conf: GeneralConfiguration):
        self._graph = nx.Graph()
        self._conf = conf

    def add_node(self, node: NodeImpl) -> None:
        # Ignore already present nodes
        if self._graph.has_node(node.id):
            return

        self._graph.add_node(node.id)

    def add_connection(self, n1: NodeImpl, n1_port_index: int, n2: NodeImpl, n2_port_index: int, net: str, connection: Optional[Connection] = None) -> Connection:
        if not n1 or not n2:
            raise Exception("Could not add connection between nonexistent nodes")

        if not connection:
            connection = ConnectionImpl()

        success = True
        result = None
        if isinstance(n1, Router):
            if isinstance(n2, Router):
                success, result = n1._connect_router(n2, connection, n2_port_index, n1_port_index)
                if success:
                    n2_port_index, n1_port_index = result
            else:
                success, result = n1._connect_node(n2, connection, n1_port_index, n2_port_index, net)
                if success:
                    n1_port_index, n2_port_index = result
        elif isinstance(n2, Router):
            success, result = n2._connect_node(n1, connection, n2_port_index, n1_port_index, net)
            if success:
                n2_port_index, n1_port_index = result
        # Direct connection
        else:
            InterfaceImpl.cast_from(n1.interfaces[n1_port_index]).connect_endpoint(
                Endpoint(n2.id, n2_port_index, n2.interfaces[n2_port_index].ip), connection)
            InterfaceImpl.cast_from(n2.interfaces[n2_port_index]).connect_endpoint(
                Endpoint(n1.id, n1_port_index, n1.interfaces[n1_port_index].ip), connection)

        if not success:
            raise Exception("Could not add connection between nodes {} and {}. Reason: {}".format(n1.id, n2.id, result))

        connection.hop = Hop(Endpoint(n1.id, n1_port_index, n1.interfaces[n1_port_index].ip),
                             Endpoint(n2.id, n2_port_index, n2.interfaces[n2_port_index].ip))
        self._graph.add_edge(n1.id, n2.id)

        return connection

    def get_node_by_id(self, id: str = "") -> Optional[NodeImpl]:
        if not id:
            return None
        else:
            return self._conf.get_object_by_id(id, NodeImpl)

    def reset(self) -> None:
        # self._nodes_by_id.clear()
        self._graph.clear()

    def resolve_ip(self, id: str, port: int) -> IPAddress:
        node = self.get_node_by_id(id)
        if not node:
            raise ValueError("Nonexistent node id provided for resolving")

        if port >= len(node.interfaces):
            raise ValueError("Nonexistent port id provided for resolving")

        return node.interfaces[port].ip

    # When creating sessions from nodes, there are two options - either nodes are connected directly, or they
    # go through a router. So correct hops are evaluated either in N-R*-N form or N-N
    # TODO: If one direction fails, session should try constructing itself in reverse order and then restructure hops
    #       so that the origin is always at the first waypoint.
    def create_session(self, owner: str, waypoints: List[Union[str, Node]], src_service: Optional[str],
                        dst_service: Optional[str], parent: Optional[Session], reverse: bool, id: Optional[str]) -> Session:
        path: List[Hop] = []
        source: NodeImpl
        session_reversed = False

        if len(waypoints) < 2:
            raise ValueError("The session path needs at least two ids")

        session_constructed = True
        for direction in ("forward", "reverse"):

            if direction == "reverse":
                if not session_constructed:
                    path.clear()
                    waypoints.reverse()
                    session_reversed = True
                    session_constructed = True
                else:
                    break

            i = 0
            while i < len(waypoints) - 1:
                # There was an error in partial session construction
                if not session_constructed:
                    break

                node0 = None
                node1 = None
                node2 = None

                def get_node_from_waypoint(self, i: int) -> Node:
                    if isinstance(waypoints[i], str):
                        node = self.get_node_by_id(waypoints[i])
                    else:
                        node = waypoints[i]
                    return node

                # Get the nodes
                node0 = get_node_from_waypoint(self, i)
                node1 = get_node_from_waypoint(self, i + 1)

                routers = []
                # N-R*-N
                if node1.type == "Router":
                    router = Router.cast_from(node1)

                    routers.append(router)
                    node2 = get_node_from_waypoint(self, i + len(routers) + 1)

                    while node2.type == "Router":
                        routers.append(Router.cast_from(node2))
                        node2 = get_node_from_waypoint(self, i + len(routers) + 1)

                    path_candidate: List[Hop] = []
                    for elements in product(node0.interfaces, node2.interfaces):
                        node0_iface = InterfaceImpl.cast_from(elements[0])
                        node2_iface = InterfaceImpl.cast_from(elements[1])

                        path_candidate.clear()

                        # Check if the next router is connected to the first node
                        if node0_iface.endpoint.id != routers[0].id:
                            continue

                        # It is, so it's a first hop
                        path_candidate.append(
                            Hop(Endpoint(NodeImpl.cast_from(node0).id, node0_iface.index, node0_iface.ip),
                                node0_iface.endpoint))

                        # Check for every router if it routes the source and destination
                        for j, r in enumerate(routers):
                            # Find if there is a forward port
                            # Ports are returned in order of priority: local IPs, remote IPs sorted by specificity (CIDR)
                            port = r.routes(node0_iface.ip, node2_iface.ip, "*")

                            # No suitable port found, try again
                            if not port:
                                break

                            path_candidate.append(Hop(Endpoint(r.id, port.index, port.ip), port.endpoint))

                        if len(path_candidate) == len(routers) + 1:
                            path.extend(path_candidate)
                            break

                    i += len(routers) + 1

                    if len(path) < i:
                        session_constructed = False
                        break
                        # raise RuntimeError("Could not find connection between {} and {} to establish a session".format(NodeImpl.cast_from(node0).id, NodeImpl.cast_from(node2).id))
                else:
                    # N-N
                    for iface in node0.interfaces:
                        node0_iface = InterfaceImpl.cast_from(iface)

                        if node0_iface.endpoint.id == NodeImpl.cast_from(node1).id:
                            path.append(Hop(Endpoint(NodeImpl.cast_from(node0).id, node0_iface.index, node0_iface.ip),
                                            node0_iface.endpoint))
                            break

                    i += 1
                    if len(path) < i:
                        session_constructed = False
                        break
                        # raise RuntimeError("Could not find connection between {} and {} to establish a session".format(NodeImpl.cast_from(node0).id, NodeImpl.cast_from(node1).id))

        if not session_constructed:
            # Sessions are always tried to be constructed in both directions, so we need to reverse the waypoints again
            waypoints.reverse()
            raise RuntimeError(
                "Could not find connection between the following waypoints to establish a session".format(waypoints))  # MYPY: Missing the parameter in string

        # If the session was constructed from the end to front, we need to reverse the path
        if session_reversed:
            path.reverse()
            for i in range(0, len(path)):
                path[i] = path[i].swap()

        return SessionImpl(owner, parent, path, src_service, dst_service, self, id)  # MYPY: Services can be None, they are optional



from abc import ABC, abstractmethod
from typing import List, Union, Optional, Tuple, Dict
from netaddr import IPAddress

from cyst.api.host.service import Service
from cyst.api.network.elements import Interface


class Node(ABC):
    """
    A network node is a main building block of simulation topology. It is modelled as a collection of services and a
    set of network interfaces.
    """

    @property
    @abstractmethod
    def type(self) -> str:
        """
        Returns the type of the node. Currently, only two types are supported - "node" and "router".

        :rtype: str
        """

    @property
    @abstractmethod
    def services(self) -> Dict[str, Service]:
        """
        Returns a collection of service instances, which are present at the node.

        :rtype: Dict[str, Service]
        """

    @property
    @abstractmethod
    def shell(self) -> Optional[Service]:
        """
        Returns a designated shell service (if there is one).

        :rtype: Optional[Service]
        """

    @property
    @abstractmethod
    def interfaces(self) -> List[Interface]:
        """
        Returns network interfaces of the node.

        :rtype: List[Interface]
        """

    @property
    @abstractmethod
    def ips(self) -> List[IPAddress]:
        """
        Returns a list of IP addresses, which are set across all interfaces of the node.

        :rtype: List[IPAddress]
        """

    @abstractmethod
    def gateway(self, ip: Union[str, IPAddress] = "") -> Optional[Tuple[IPAddress, int]]:
        """
        Calculates, which gateway would be used by the node, if a message was sent to a given destination IP address.
        The gateway is selected according to the routing rules.

        :param ip: The destination IP address.
        :type ip: Union[str, IPAddress]

        :return: A gateway IP address together with port index (if one is found).
        :rtype: Optional[Tuple[IPAddress, int]]
        """



from __future__ import annotations

import asyncio
import inspect
import sys

from heapq import heappush
from netaddr import IPAddress
from typing import TYPE_CHECKING, Optional, Any, Union, Dict, List

from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.message import Request, Response, Status, Message, MessageType, StatusValue, StatusOrigin, \
    ComponentState, Signal
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.logic.access import Authorization, AuthenticationTarget, AuthenticationToken
from cyst.api.logic.action import Action, ActionType
from cyst.api.logic.metadata import Metadata
from cyst.api.network.node import Node
from cyst.api.network.session import Session
from cyst.api.host.service import ActiveService

from cyst.core.logic.action import ActionImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


class EnvironmentMessagingImpl(EnvironmentMessaging):
    def __init__(self, env: _Environment):
        self._env = env

    def send_message(self, message: Message, delay: int = 0) -> None:
        _send_message(self._env, message, delay)

    def create_request(self, dst_ip: Union[str, IPAddress], dst_service: str = "", action: Optional[Action] = None,
                       session: Optional[Session] = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None,
                       original_request: Optional[Request] = None) -> Request:
        return self._env.platform.messaging.create_request(dst_ip, dst_service, action, session, auth, original_request)

    def create_response(self, request: Request, status: Status, content: Optional[Any] = None,
                        session: Optional[Session] = None,
                        auth: Optional[Union[Authorization, AuthenticationTarget]] = None,
                        original_response: Optional[Response] = None):
        return self._env.platform.messaging.create_response(request, status, content, session, auth, original_response)

    def create_signal(self, signal_origin: str, state: ComponentState, effect_origin: str,
                      effect_message: Optional[int] = None, effect_description: str = "",
                      effect_parameters: Optional[Dict[str, Any]] = None) -> Signal:
        return self._env.platform.messaging.create_signal(signal_origin, state, effect_origin, effect_message,
                                                          effect_description, effect_parameters)

    def open_session(self, request: Request, reverse_direction: bool = False) -> Session:
        return self._env.platform.messaging.open_session(request, reverse_direction)


# ----------------------------------------------------------------------------------------------------------------------
def extract_metadata_action(action: Action, action_list: List[Action]):
    if not action.components:
        action_list.append(action)
    else:
        for c in action.components:
            extract_metadata_action(c, action_list)


def _send_message(self: _Environment, message: Message, delay: int = 0) -> None:

    if isinstance(message, Request):
        # This is a hack that enables getting the caller of the send_message() function.
        # It is ugly, but better than alternatives.

        # Caller frame is up two places
        caller_frame = inspect.currentframe().f_back.f_back
        caller = caller_frame.f_locals["self"]

        if caller and isinstance(caller, ActiveService):
            caller_id = self._service_store.get_active_service_id(id(caller))
            message.platform_specific["caller_id"] = caller_id
            if not caller_id in self._action_counts:
                self._action_counts[caller_id] = 0

    if message.type == MessageType.REQUEST or message.type == MessageType.RESPONSE:
        self._data_store.add_message(message)
    elif message.type == MessageType.SIGNAL:
        self._data_store.add_signal(message)

    # I would much rather check by the MessageType, but then the type inspection would not work down the line :-/
    if isinstance(message, Request) or isinstance(message, Response):

        # ------------------------------------------------------------------------------------------------------------------
        # Processing depends on action type
        # ------------------------------------------------------------------------------------------------------------------
        action_type = ActionImpl.cast_from(message.action).type

        # ------------------------------------------------------------------------------------------------------------------
        # Composite actions
        if action_type == ActionType.COMPOSITE:
            # Request are sent to composite action manager to process
            # Responses are processed outside the messaging interface and within the main loop
            if message.type == MessageType.REQUEST:
                self._cam.execute_request(message, delay)
                # Top-level composite actions counts towards the action limit
                caller_id = message.platform_specific["caller_id"]
                action_count = self._action_counts[caller_id] + 1
                self._action_counts[caller_id] = action_count

                if action_count > self._highest_action_count:
                    self._highest_action_count = action_count
                    self._highest_action_actor = caller_id

                action_record = ActionModel.from_request(self.infrastructure.statistics.run_id, message)
                # Composite messages do not have an assigned IP, because they are (in theory) executed locally. So, we
                # select one "random" IP from nodes IP list and use it as a source IP.
                node_id, _ = caller_id.split(".")
                node: Node = self._general_configuration.get_object_by_id(node_id, Node)
                if node.ips:
                    action_record.src_ip = str(node.ips[0])
                self._active_actions[message.id] = action_record
            else:
                # Send it to the platform.
                self._platform.messaging.send_message(message, delay)

        # ------------------------------------------------------------------------------------------------------------------
        # Direct actions
        elif action_type == ActionType.DIRECT:
            # Only non-composite-related messages count towards the action limit
            if message.type == MessageType.REQUEST and not self._cam.is_composite(message.id):
                caller_id = message.platform_specific["caller_id"]
                action_count = self._action_counts[caller_id] + 1
                self._action_counts[caller_id] = action_count

                if action_count > self._highest_action_count:
                    self._highest_action_count = action_count
                    self._highest_action_actor = caller_id

                self._active_actions[message.id] = ActionModel.from_request(self.infrastructure.statistics.run_id,
                                                                            message)

            # --------------------------------------------------------------------------------------------------------------
            # Call the behavioral model to add components to direct actions
            # We ignore actions without behavioral models, such as direct agent-agent actions.
            if message.action.namespace in self._behavioral_models:
                message.action.components.extend(self._behavioral_models[message.action.namespace].action_components(message))

            # --------------------------------------------------------------------------------------------------------------
            # Enrich the message with metadata. The rule of thumb is that actions with components get the metadata from
            # them. Otherwise, their metadata provider is queried.
            action_queue = []
            extract_metadata_action(message.action, action_queue)

            message_metadata = Metadata()
            message_metadata.flows = []

            # TODO: This is only temporary and probably a subject to changes, because of many undefined corner cases
            for action in action_queue:
                for namespace, provider in self._metadata_providers.items():
                    if action.id.startswith(namespace):
                        metadata = provider.get_metadata(action, message)
                        # TODO: Currently we are only considering flows
                        if metadata.flows:
                            message_metadata.flows.extend(metadata.flows)

            message.set_metadata(message_metadata)

            # --------------------------------------------------------------------------------------------------------------
            # Send it to the platform.
            self._platform.messaging.send_message(message, delay)

        # ------------------------------------------------------------------------------------------------------------------
        # Component actions
        # Component actions should never travel through the systems as top-level actions of a message
        else:
            if message.type == MessageType.RESPONSE:
                raise RuntimeError("Component action ended in a response. This indicates error in logic somewhere in the pipeline before.")
            else:
                r = self.messaging.create_response(message.cast_to(Request), status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                   content="Component action in response. This should not happen.",
                                                   session=message.session, auth=message.auth)
                self.messaging.send_message(r)
                return

    # Pause on request is called regardless of action type
    if message.type is MessageType.REQUEST and message.platform_specific["caller_id"] in self._pause_on_request:
        self._pause = True



from cyst.api.environment.configuration import RuntimeConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import ServiceStore, DataStore


class EnvironmentInfrastructureImpl(EnvironmentInfrastructure):

    def __init__(self, runtime_configuration: RuntimeConfiguration, data_store: DataStore, service_store: ServiceStore,
                 statistics: Statistics):
        self._runtime_configuration = runtime_configuration
        self._data_store = data_store
        self._service_store = service_store
        self._statistics = statistics

    @property
    def statistics(self) -> Statistics:
        return self._statistics

    @property
    def data_store(self) -> DataStore:
        return self._data_store

    @property
    def service_store(self) -> ServiceStore:
        return self._service_store

    @property
    def runtime_configuration(self) -> RuntimeConfiguration:
        return self._runtime_configuration



import enum
import time
import importlib

import flags
import jsonpickle

from typing import Any

import semver
from netaddr import IPAddress, IPNetwork

import netaddr

from cyst.api.configuration.logic.access import AuthorizationDomainType
from cyst.api.configuration.network.elements import ConnectionConfig
from cyst.api.environment.configuration import ServiceParameter
from cyst.api.configuration.configuration import ConfigParameterValueType, ConfigParameterGroupType
from cyst.api.environment.control import EnvironmentState
from cyst.api.environment.message import MessageType, StatusOrigin, StatusValue, StatusDetail
from cyst.api.host.service import ServiceState
from cyst.api.logic.access import AccessLevel, AuthenticationTokenType, AuthenticationTokenSecurity, AuthenticationProviderType
from cyst.api.logic.action import ActionParameterDomainType, ActionParameterType
from cyst.api.logic.exploit import ExploitCategory, ExploitLocality, ExploitParameterType
from cyst.api.logic.metadata import TCPFlags, Protocol, FlowDirection
from cyst.api.network.firewall import FirewallPolicy, FirewallChainType

# ----------------------------------------------------------------------------------------------------------------------
# Custom serialization handlers
# ----------------------------------------------------------------------------------------------------------------------


@jsonpickle.handlers.register(IPNetwork)
class IPNetworkSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj, data):
        data["_value"] = str(obj)
        return data

    def restore(self, obj):
        return IPNetwork(obj["_value"])


@jsonpickle.handlers.register(IPAddress)
class IPAddressSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj, data):
        data["_value"] = str(obj)
        return data

    def restore(self, obj):
        return IPAddress(obj["_value"])


@jsonpickle.handlers.register(time.struct_time)
class TimeSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj: time.struct_time, data):
        data["_value"] = time.mktime(obj)
        return data

    def restore(self, obj):
        return time.localtime(obj["_value"])


@jsonpickle.handlers.register(enum.Enum, base=True)
class EnumSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj: enum.Enum, data):
        data["_value"] = obj.name
        return data

    def restore(self, obj):
        enum_name = obj["py/object"].split(".")[-1]
        return globals()[enum_name][obj["_value"]]


@jsonpickle.handlers.register(flags.Flags, base=True)
class FlagsSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj: flags.Flags, data):
        data["_value"] = obj.to_simple_str()
        return data

    def restore(self, obj):
        flag_name = obj["py/object"].split(".")[-1]
        return globals()[flag_name].from_simple_str(obj["_value"])


@jsonpickle.handlers.register(semver.VersionInfo, base=True)
class VersionInfoSerializer(jsonpickle.handlers.BaseHandler):

    def flatten(self, obj: semver.VersionInfo, data):
        data["_value"] = str(obj)
        return data

    def restore(self, obj):
        return semver.VersionInfo.parse(obj["_value"])


# Environment
class Serializer:
    @staticmethod
    def serialize(obj: Any) -> str:
        return jsonpickle.encode(obj, make_refs=True, indent=2, keys=True)

    @staticmethod
    def deserialize(state: str) -> Any:
        obj = jsonpickle.decode(state, keys=True)
        return obj



from cyst.api.environment.environment import Environment

e = Environment.create()
e.control.init()
e.control.run()
e.control.commit()

stats = e.infrastructure.statistics

print(f"Run id: {stats.run_id}\nStart time real: {stats.start_time_real}\n"
      f"End time real: {stats.end_time_real}\nDuration virtual: {stats.end_time_virtual}")



import uuid

from typing import Optional

from cyst.api.logic.data import Data


class DataImpl(Data):
    def __init__(self, id: Optional[str], owner: str, path: str = "", description: str = ""):
        if id:
            self._id = id
        else:
            self._id = str(uuid.uuid4())
        self._owner = owner
        self._path = path
        self._description = description

    @property
    def id(self) -> str:
        return self._id

    @property
    def owner(self) -> str:
        return self._owner

    @property
    def path(self) -> str:
        return self._path

    @property
    def description(self) -> str:
        return self._description



import logging
import asyncio
import time

from datetime import datetime
from heapq import heappop
from typing import List, Tuple, Union, Optional, Any, Callable

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.environment.clock import Clock
from cyst.api.environment.configuration import EnvironmentConfiguration, GeneralConfiguration, NodeConfiguration, \
    ServiceConfiguration, NetworkConfiguration, ExploitConfiguration, AccessConfiguration, ActionConfiguration, \
    PhysicalConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.message import Message, MessageType, Timeout
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.platform import Platform, PlatformDescription
from cyst.api.environment.platform_interface import PlatformInterface
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.host.service import ActiveService
from cyst.api.network.node import Node
from cyst.api.network.session import Session

from cyst.platform.host.service import Service, ServiceImpl
from cyst.platform.environment.configuration_access import AccessConfigurationImpl
from cyst.platform.environment.configuration_general import GeneralConfigurationImpl
from cyst.platform.environment.configuration_network import NetworkConfigurationImpl
from cyst.platform.environment.configuration_node import NodeConfigurationImpl
from cyst.platform.environment.configuration_service import ServiceConfigurationImpl
from cyst.platform.environment.configurator import Configurator
from cyst.platform.environment.message import TimeoutImpl
from cyst.platform.environment.environment_messaging import EnvironmentMessagingImpl
from cyst.platform.network.network import Network
from cyst.platform.network.session import SessionImpl


class CYSTPlatform(Platform, EnvironmentConfiguration, Clock):
    def __init__(self, platform_interface: PlatformInterface, general_configuration: GeneralConfiguration,
                 resources: EnvironmentResources, action_configuration: ActionConfiguration,
                 exploit_configuration: ExploitConfiguration, infrastructure: EnvironmentInfrastructure,
                 physical_configuration: PhysicalConfiguration, platform_type: PlatformType):
        self._platform_interface = platform_interface
        self._resources = resources
        self._action_configuration = action_configuration
        self._exploit_configuration = exploit_configuration
        self._physical_configuration = physical_configuration
        self._infrastructure = infrastructure
        self._platform_type = platform_type
        self._real_time_wait_factor = 0.1

        self._message_log = logging.getLogger("messaging")

        self._time = 0.0
        self._init_time = 0.0

        self._message_queue: List[Tuple[int, int, Message]] = []
        self._execute_queue: List[Tuple[int, int, Message]] = []

        self._messages_processing = 0

        self._general_configuration = GeneralConfigurationImpl(self, general_configuration)
        self._access_configuration = AccessConfigurationImpl(self)
        self._network_configuration = NetworkConfigurationImpl(self)
        self._node_configuration = NodeConfigurationImpl(self)
        self._service_configuration = ServiceConfigurationImpl(self)

        self._network = Network(self._general_configuration)
        self._sessions_to_add: List[Tuple[str, List[Union[str, Node]], Optional[str], Optional[str], Optional[Session], bool, Optional[str]]] = []

        self._environment_messaging = EnvironmentMessagingImpl(self)

    def init(self) -> bool:
        for session in self._sessions_to_add:
            owner = session[0]
            waypoints = session[1]
            src_service = session[2]
            dst_service = session[3]
            parent = session[4]
            reverse = session[5]
            id = session[6]

            s: SessionImpl = SessionImpl.cast_from(self._network.create_session(owner, waypoints, src_service, dst_service, parent, reverse, id))

            # Add sessions to services
            src_service = ServiceImpl.cast_from(self._general_configuration.get_object_by_id(f"{s.startpoint.id}.{src_service}", Service))
            dst_service = ServiceImpl.cast_from(self._general_configuration.get_object_by_id(f"{s.endpoint.id}.{dst_service}", Service))

            src_service.sessions[s.id] = s
            dst_service.sessions[s.id] = s

        self._init_time = time.time()
        return True

    def terminate(self) -> bool:
        pass

    def configure(self, *config_item: ConfigItem) -> 'Platform':
        self._general_configuration.configure(*config_item)
        return self

    # ------------------------------------------------------------------------------------------------------------------
    # Configuration
    # ------------------------------------------------------------------------------------------------------------------
    @property
    def configuration(self) -> EnvironmentConfiguration:
        return self

    @property
    def general(self) -> GeneralConfiguration:
        return self._general_configuration

    @property
    def node(self) -> NodeConfiguration:
        return self._node_configuration

    @property
    def service(self) -> ServiceConfiguration:
        return self._service_configuration

    @property
    def network(self) -> NetworkConfiguration:
        return self._network_configuration

    @property
    def exploit(self) -> ExploitConfiguration:
        return self._exploit_configuration

    @property
    def action(self) -> ActionConfiguration:
        return self._action_configuration

    @property
    def access(self) -> AccessConfiguration:
        return self._access_configuration

    @property
    def physical(self) -> PhysicalConfiguration:
        return self._physical_configuration

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def messaging(self) -> EnvironmentMessaging:
        return self._environment_messaging

    @property
    def policy(self) -> EnvironmentPolicy:
        pass

    @property
    def clock(self) -> Clock:
        return self

    # ------------------------------------------------------------------------------------------------------------------
    # Clock interface
    def current_time(self) -> float:
        if self._platform_type == PlatformType.REAL_TIME:
            self._time = time.time()

        return self._time

    def real_time(self) -> datetime:
        if self._platform_type == PlatformType.REAL_TIME:
            self._time = time.time()
            return datetime.fromtimestamp(self._time)
        else:
            return datetime.fromtimestamp(self._init_time + self._time)

    def timeout(self, callback: Union[ActiveService, Callable[[Timeout], None]], delay: float, parameter: Any = None) -> None:
        timeout = TimeoutImpl(callback, self.current_time(), delay, parameter)
        self._environment_messaging.send_message(timeout, int(delay))

    # ------------------------------------------------------------------------------------------------------------------
    def _finish_message_processing(self, task: asyncio.Task):
        self._messages_processing -= 1

    # ------------------------------------------------------------------------------------------------------------------
    async def process(self, time_advance: int) -> bool:

        if self._platform_type == PlatformType.REAL_TIME:
            self._time = time.time()

        have_something_to_do = bool(self._message_queue) or bool(self._execute_queue) or self._messages_processing > 0
        time_jump = 0

        # Message-passing tasks
        if self._message_queue:
            next_time = self._message_queue[0][0]

            delta = next_time - self._time
            if time_jump == 0 or delta < time_jump:
                time_jump = delta

        # Request execution tasks
        if self._execute_queue:
            next_time = self._execute_queue[0][0]

            delta = next_time - self._time
            if time_jump == 0 or delta < time_jump:
                time_jump = delta

        # Messages hopping
        # If hops are asynchronously processed, then we must not move the time forward, but we must enable processing
        # of another messages and messages hops that may arise from them.
        if self._messages_processing > 0:
            time_jump = 0

        if self._platform_type == PlatformType.SIMULATED_TIME:
            # If there is nothing to do, just jump simulated time as asked
            if not have_something_to_do:
                self._time += time_advance
                return False
            else:
                # If there is something to do, but it is further than the environment requested, we just move the clock and
                # do nothing
                if time_advance > 0 and time_jump > time_advance:
                    self._time += time_advance
                    return True
                # It is sooner than the environment requested, let's do it and proceed with the rest of the code
                else:
                    self._time += time_jump
        else:
            # We do not jump in time more than ordered
            if time_advance > 0 and time_jump > time_advance:
                time_jump = time_advance

            # We do not jump in time more than the wait factor
            if time_jump > self._real_time_wait_factor:
               time_jump = self._real_time_wait_factor

            # We jump the minimum needed
            if time_jump > 0:
                # Not async - we really want to let it wait, because we know that nothing can happen in the meantime
                # await asyncio.sleep(time_jump)
                time.sleep(time_jump)

            # And signal that we do not have anything to do
            if not have_something_to_do:
                return False

        # --------------------------------------------------------------------------------------------------------------
        # Task processing
        messages_to_process = []

        if self._message_queue:
            next_time = self._message_queue[0][0]
            while next_time <= self._time:
                messages_to_process.append(heappop(self._message_queue)[2])
                if self._message_queue:
                    next_time = self._message_queue[0][0]
                else:
                    break

        for message in messages_to_process:
            if message.type == MessageType.TIMEOUT:
                # Yay!
                timeout = TimeoutImpl.cast_from(message.cast_to(Timeout))  # type:ignore #MYPY: Probably an issue with mypy, requires creation of helper class
                timeout.callback(message)
            else:
                t: asyncio.Task = asyncio.get_running_loop().create_task(self._environment_messaging.message_hop(message))
                t.add_done_callback(self._finish_message_processing)
                self._messages_processing += 1

        tasks_to_execute = []

        if self._execute_queue:
            next_time = self._execute_queue[0][0]
            while next_time <= self._time:
                tasks_to_execute.append(heappop(self._execute_queue)[2])
                if self._execute_queue:
                    next_time = self._execute_queue[0][0]
                else:
                    break

        for task in tasks_to_execute:
            t: asyncio.Task = asyncio.get_running_loop().create_task(self._environment_messaging.message_process(task))
            t.add_done_callback(self._finish_message_processing)
            self._messages_processing += 1

        # Yield processing to the event loop if there are messages being processed
        if self._messages_processing > 0:
            await asyncio.sleep(0)

        return True


def create_simulated_time_platform(platform_interface: PlatformInterface, general_configuration: GeneralConfiguration,
                                   resources: EnvironmentResources, action_configuration: ActionConfiguration,
                                   exploit_configuration: ExploitConfiguration, physical_configuration: PhysicalConfiguration,
                                   infrastructure: EnvironmentInfrastructure) -> CYSTPlatform:
    p = CYSTPlatform(platform_interface, general_configuration, resources, action_configuration, exploit_configuration,
                     infrastructure, physical_configuration, PlatformType.SIMULATED_TIME)
    return p

def create_real_time_platform(platform_interface: PlatformInterface, general_configuration: GeneralConfiguration,
                              resources: EnvironmentResources, action_configuration: ActionConfiguration,
                              exploit_configuration: ExploitConfiguration, physical_configuration: PhysicalConfiguration,
                              infrastructure: EnvironmentInfrastructure) -> CYSTPlatform:
    p = CYSTPlatform(platform_interface, general_configuration, resources, action_configuration, exploit_configuration,
                     infrastructure, physical_configuration, PlatformType.REAL_TIME)
    return p


simulated_time_platform_description = PlatformDescription(
    specification=PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"),
    description="A platform implementation for the CYST simulation engine, which is using simulated discrete time.",
    creation_fn=create_simulated_time_platform
)

real_time_platform_description = PlatformDescription(
    specification=PlatformSpecification(PlatformType.REAL_TIME, "CYST"),
    description="A platform implementation for the CYST simulation engine, which is using real time.",
    creation_fn=create_real_time_platform
)



from abc import ABC, abstractmethod

from typing import List, Tuple, Union, Optional

from cyst.api.host.service import Service
from cyst.api.logic.access import AccessLevel, AccessScheme, Authorization
from cyst.api.network.node import Node


class EnvironmentPolicy(ABC):
    """
    This environment enables handling of authentication and authorization.

    Warning:
        This interface will be gradually phased off in favor of a new auth framework.
    """

    @abstractmethod
    def create_authorization(self, identity: str, nodes: List[Union[str, Node]], services: List[Union[str, Service]],
                             access_level: AccessLevel, id: str, token: Optional[str] = None) -> Authorization:
        """
        Creates an authorization token, which can be used to get access to service resources.

        :param identity: An identity of a user this token belongs to. Usually the identity of a resource and the
            identity of a token must match to get authorized.
        :type identity: str

        :param nodes: A list of nodes this authorization token can be used on. Either a node instance or a node ID
            can be used.
        :type nodes: List[Union[str, Node]]

        :param services: A list of services this authorization token can be used on. Either a service instance or a
            service ID can be used.
        :type services: List[Union[str, Service]]

        :param access_level: The access level of the service.
        :type access_level: AccessLevel

        :param id: An ID of the authorization token. This ID has to be unique across simulation to enable correct
            evaluation of access rights.
        :type id: str

        :param token: An optional data payload related to the authorization token. Mostly useless now, as the payload
            is more important for authentication token, that were established within the new auth framework.
        :type token: Optional[str]

        :return: An authorization token
        """

    @abstractmethod
    def get_authorizations(self, node: Union[str, Node], service: str, access_level: AccessLevel = AccessLevel.NONE) -> List[Authorization]:
        """
        Gets authorization templates for a given service on a given node with a specified access level. Note that these
        are _templates_ and not actual authorizations. For more information, see description of the auth framework.

        :param node: An instance or an ID of the node.
        :type node: Union[str, Node]

        :param service: A name of the service.
        :type service: str

        :param access_level: An access level that the authorization should have.
        :type access_level: AccessLevel

        :return: A list of authorization templates.
        """

    @abstractmethod
    def decide(self, node: Union[str, Node], service: str, access_level: AccessLevel, authorization: Authorization) -> Tuple[bool, str]:
        """
        Evaluates, whether an authorization is applicable for a given service, on a given node and with given
        access level.

        :param node: An instance or an ID of the node.
        :type node: Union[str, Node]

        :param service: A name of the service.
        :type service: str

        :param access_level: An access level that the authorization should have.
        :type access_level: AccessLevel

        :param authorization: The authorization that is being tested.
        :type authorization: Authorization

        :return: An information if the authorization is applicable and if not, also the reason why.

            :Applicable authorization: (True, "")
            :Wrong authorization: (False, "Reason")

        """

    @abstractmethod
    def get_schemas(self, node: Union[str, Node], service: str) -> List[AccessScheme]:
        """
        schemas under which the service is accessible
        TODO
        """

    @abstractmethod
    def get_nodes(self, authorization: Authorization) -> List[str]:
        """
        Gets the IDs of nodes that this authorization pertain to. This function exists, because the nodes list is not
        accessible via the :class:`cyst.api.logic.access.Authorization` interface.

        :param authorization: An instance of the authorization.
        :type authorization: Authorization
        """

    @abstractmethod
    def get_services(self, authorization: Authorization) -> List[str]:
        """
        Gets the IDs of services that this authorization pertain to. This function exists, because the services list is
        not  accessible via the :class:`cyst.api.logic.access.Authorization` interface.

        :param authorization: An instance of the authorization.
        :type authorization: Authorization
        """

    @abstractmethod
    def get_access_level(self, authorization: Authorization) -> AccessLevel:
        """
        Gets the access level of the authorization. TODO: I have no idea, why this function exists, because access level
        is accessible via the Authorization interface.

        :param authorization: An instance of the authorization.
        :type authorization: Authorization
        """



import uuid

from abc import ABC, abstractmethod
from enum import Enum
from typing import Tuple


class EnvironmentState(Enum):
    """
    State of the environment

    State transition diagram:

    ::

                              reset()                 terminate()
                             ┌────────── TERMINATED ◄───────────┐
                             │               ▲                  │
                             │    terminate()│                  │
                             ▼               │    ──pause()──►  │
        CREATED ──init()─► INIT ──run()─► RUNNING             PAUSED
                             ▲               │    ◄──run()───
                             │               │
                             │               ▼
                             └────────── FINISHED
                              reset()

    Possible values:
        :CREATED: The environment was jus created.
        :INIT: The environment was initialized and is ready to run.
        :RUNNING: The environment is currently running a simulation.
        :PAUSED: The environment is paused. If unpaused, the messages on stack will be sent and simulation resumes.
        :FINISHED: The environment finished a simulation run. This means either no more messages on stack, or running
            past the specified time.
        :TERMINATED: The environment was forcefully stopped. It cannot be resumed, but its internal state can be
            investigated.

    """
    CREATED = -1,
    INIT = 0,
    RUNNING = 1,
    PAUSED = 2,
    FINISHED = 3,
    TERMINATED = 4


class EnvironmentControl(ABC):
    """
    EnvironmentControl provides mechanisms to control the execution of actions within the simulation environment.

    Availability:
        :Available: creator
        :Hidden:  agents, models
    """

    @property
    @abstractmethod
    def state(self) -> EnvironmentState:
        """ Provides the current state of the environment.

        :return: State of the environment
        :rtype: EnvironmentState
        """

    @abstractmethod
    def init(self) -> Tuple[bool, EnvironmentState]:
        """ Initializes the environment for the first time. The environment must be in the CREATED state.
        The following invocations do nothing and silently return true.

        :return: A tuple indicating, whether the operation was successful and which state the environment ended in.
        """

    @abstractmethod
    def commit(self) -> None:
        """ Stores the information of the currently executed run into the data store. This can only be executed from
        the FINISHED or TERMINATED state.

        :return: None
        """

    @abstractmethod
    def reset(self, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
        """ Resets the environment for another run. Only a previously FINISHED or TERMINATED run can be reset.

        :param run_id: The unique id of the current run. If a non-unique id is selected, it may produced unwanted
            results when saving the data to a data store.
        :type run_id: str

        :return: A tuple indicating, whether the operation was successful and which state the environment ended in.
        """

    @abstractmethod
    def run(self) -> Tuple[bool, EnvironmentState]:
        """ Starts or resumes the message processing in the current run. If the environment is in the INIT state, it
        activates the active services. If it is in INIT or PAUSED state, it begins message processing and transitions
        into the RUNNING state.

        :return: A tuple indicating, whether the operation was successful and which state the environment ended in.
        """

    @abstractmethod
    def pause(self) -> Tuple[bool, EnvironmentState]:
        """ Invokes an explicit transition into the PAUSED state and temporarily halts the message processing. Can only
        be applied in the running state.

        :return: A tuple indicating, whether the operation was successful and which state the environment ended in.
        """

    @abstractmethod
    def terminate(self) -> Tuple[bool, EnvironmentState]:
        """ Halts the message processing and transitions the environment into the TERMINATED state. From this state
        the environment cannot be re-run.

        :return: A tuple indicating, whether the operation was successful and which state the environment ended in.
        """

    @abstractmethod
    def add_pause_on_request(self, id: str) -> None:
        """ Adds an explicit interrupt to message processing, whenever a service sends a request. Transitions the
        environment into the PAUSED state. Used mainly in tests to break from the run().

        :param id: A fully qualified id of a service, i.e., also containing the node id.

        :return: None
        """

    @abstractmethod
    def remove_pause_on_request(self, id: str) -> None:
        """ Removes an explicit interrupt to message processing, whenever a service sends a request.

        :param id: A fully qualified id of a service, i.e., also containing the node id.

        :return: None
        """

    @abstractmethod
    def add_pause_on_response(self, id: str) -> None:
        """ Adds an explicit interrupt to message processing, whenever a service receives a response. Transitions the
        environment into the PAUSED state. Used mainly in tests to break from the run().

        :param id: A fully qualified id of a service, i.e., also containing the node id.

        :return: None
        """

    @abstractmethod
    def remove_pause_on_response(self, id: str) -> None:
        """ Removes an explicit interrupt to message processing, whenever a service sends a request.

        :param id: A fully qualified id of a service, i.e., also containing the node id.

        :return: None
        """

    @abstractmethod
    def snapshot_save(self) -> str:
        """ Returns a complete state of the environment, from which it can be restored.

        :return: String representation of the environment
        """

    @abstractmethod
    def snapshot_load(self, state: str) -> None:
        """ Attempts to restore the environment into the state described by the state string.

        :param state: State representation obtained by snapshot_save function. Currently, the snapshots are not
                      guaranteed to work across versions.
        :type state: str

        :return: None
        """

    @abstractmethod
    def transaction_start(self) -> Tuple[int, int, str]:
        """ Initiates a new transaction, which sets a point in time to which the environment can be reverted. The
        transaction has to begin as a first thing in a given time slot, i.e., if there was already a message processed
        in the time slot, the transaction will be set to start in the next time slot.

        Multiple calls to transaction_start() at the same time slot start only one transaction.

        Transactions can be interleaved. In case of a rollback, the state always returns to the beginning of
        transaction that was rolled back.

        Transactions do not necessarily modify the state of active services, it is up to their authors, if they want
        to preserve, e.g., learned behavior across transactions.

        :return: A tuple containing transaction_id of started transaction, the time when the transaction starts, and
                 additional information, if there is any.
        """

    @abstractmethod
    def transaction_commit(self, transaction_id: int) -> Tuple[bool, str]:
        """
        Finalizes a transaction. Such transaction is removed from the system and cannot be rolled back.

        Any transaction can be committed by anyone. There is no validation of the caller.

        :param transaction_id: The id of a transaction to commit.
        :type transaction_id: int

        :return: A tuple indicating whether the transaction was successfully committed and providing additional
                 information if present.
        """

    @abstractmethod
    def transaction_rollback(self, transaction_id: int) -> Tuple[bool, str]:
        """
        Returns the system into the state at the beginning of a transaction. After the transaction is rolled back, it is
        removed from the system.

        Any transaction can be rolled back by anyone. There is no validation of a caller.

        :param transaction_id: The ID of the transaction to roll back.
        :type transaction_id: int

        :return: A tuple indicating whether a rollback was successful and providing additional information in present.
        """



from abc import ABC, abstractmethod
from enum import Enum
from typing import Union, Optional, Type
from urllib.parse import ParseResult


from cyst.api.host.service import ActiveService

class ResourcePersistence(Enum):
    """
    Resources can be created in one of the two modes - transient and persistent. In the transient mode, they are
    opened and closed for each operation on them. In the persistent mode, they get opened on the first operation
    and must be closed via the :func:`release_resource` function.

    Possible values:
        :TRANSIENT: Open/close on each operation.
        :PERSISTENT: Stay opened after the first operation.

    """
    TRANSIENT = 0
    PERSISTENT = 1


class Resource(ABC):
    """
    Resource interface represents an interface that is user-facing. A user does not interact with the resource directly,
    but is instead using the :class:`ExternalResources` interface. Therefore, this interface provides only some minimal
    information about the resource.
    """
    @property
    @abstractmethod
    def path(self) -> str:
        """
        The URL of the resource. Technically, many different resources can share the same URL and be differentiated just
        by the parameters that are provided with particular operations on the :class:`ExternalResources` interface.
        Therefore, this property is only informative.
        """

    @property
    @abstractmethod
    def persistence(self) -> ResourcePersistence:
        """
        The persistence setting of the resource.
        """


class ResourceImpl(Resource, ABC):
    """
    This interface represents an implementation of a resource that is used by the :class:`ExternalResource` interface.
    Any new or custom resources that are added to the system must inherit from both :class:`Resource` and
    :class:`ResourceImpl`.
    """
    @abstractmethod
    def init(self, path: ParseResult, params: Optional[dict[str, str]] = None, persistence: ResourcePersistence = ResourcePersistence.TRANSIENT) -> bool:
        """
        This function is called when the resource is initialized after being created.

        :param path: The result of urllib.parse() call on the URL. The scheme is always guaranteed to be the same as
            the one used when this resource was registered.
        :type path: ParseResult

        :param params: Arbitrary parameters that may be required for resource initialization. If the resource is created
            implicitly via the ExternalResources interface, these parameters are shared with the call to
            :func:`send`/:func:`receive`.
        :type params: Optional[dict[str, str]]

        :param persistence: Persistence of a result. Implicitly created resources are always transient.
        :type persistence: ResourcePersistence

        :return: True if the resource was successfully created and False otherwise.
        """

    @abstractmethod
    def open(self) -> None:
        """
        This function is called prior to send/receive functions. It should prepare the resource for interactions.

        :return: None
        """

    @abstractmethod
    def close(self) -> None:
        """
        This function closes the resource. Any interaction after closing should fail. Close is either called immediately
        after :func:`send`/:func:`receive` operation for transient resources, or called when the resource is released
        for persistent ones.

        :return: None
        """

    @abstractmethod
    async def send(self, data: str, params: Optional[dict[str, str]] = None) -> int:
        """
        This function sends the data to the resource, e.g., writes to the socket, writes to a file, or inserts into
        a database. Due to its async nature, you should utilize the async I/O operations, as the sync ones can
        disturb the execution of the rest of the processing.

        :param data: The data that should be written.
        :type data: str

        :param params: Arbitrary parameters that modify the send operation. If the resource was implicitly created,
            these parameters may contain ones that are related to resource creation.
        :type params: Optional[dict[str, str]]

        :return: The number of characters that were actually written, even though the expectation here is that this
            function sends all the data or fails.
        """

    @abstractmethod
    async def receive(self, params: Optional[dict[str, str]] = None) -> Optional[str]:
        """
        This function read the data from the resource. Due to its async nature, you should utilize the async I/O
        operations, as the sync ones can disturb the execution of the rest of the processing.

        :param params: Arbitrary parameters that modify the receive operation. If the resource was implicitly created,
            these parameters may contain ones that are related to resource creation.
        :type params: Optional[dict[str, str]]

        :return: The data received from the resource or None if nothing was available or expected, e.g., HEAD HTTP
            request.
        """


class ExternalResources(ABC):
    """
    External resources represent any resources that are not part of the simulation/emulation runs, such as files,
    REST APIs, databases, etc. To maintain a sensible and consistent relation to underlying platforms and their time,
    all such resources must be handled through this interface.

    .. warning::
        Due to technical reasons related to dependence on an underlying platform, ExternalResources can be used only
        in the init/post-init state, including the attempts to register custom resources. That is, resources cannot be
        managed from the __init__ methods of behavioral models, platforms, or services.

    """

    @abstractmethod
    def register_resource(self, scheme: str, resource: Union[Type, ResourceImpl]) -> bool:
        """
        Register a given resource for given scheme.

        :param scheme: URL scheme, i.e., the part before ://.
        :type scheme: str

        :param resource: The resource can either be provided as a type that implements the Resource and ResourceImpl
            interfaces, or it can be an object instance. If a type is provided then a new instance of the resource is
            created for each call of :func:`create_resource`, including the implicit call in sending and fetching
            function. If a resource instance is provided then this resource acts as a singleton and multiple interleaved
            :func:`open` and :func:`close` calls can happen and the resource must be prepared to handle it gracefully.
        :type resource: Union[Type, ResourceImpl]

        :return: True if the resource was successfully registered. False otherwise.
        """

    @abstractmethod
    def create_resource(self, path: str, params: Optional[dict[str, str]] = None, persistence: ResourcePersistence = ResourcePersistence.TRANSIENT) -> Resource:
        """
        Creates an instance of a resource with a given path.

        :param path: The URL of the resource. Full URL with scheme, e.g., file:// must be supplied to let system
            choose the correct one.
        :type path: str

        :param params: Arbitrary set of parameters that will be passed to the resource's init function. If the resource
            is a singleton (cf. :func:`register_resource`) then these parameters are ignored.
        :type params: Optional[dict[str, str]]

        :param persistence: The persistence of a resource. The :func:`create_resource` function is usually called
            implicitly from the other functions, which set the persistence to transient. However, if you do not want
            the resource to be created and destroyed after each call, set the persistence to
            ResourcePersistence.PERSISTENT.
        :type persistence: ResourcePersistence

        :return: An instance of a resource. Otherwise, an exception is thrown.
        """

    @abstractmethod
    def release_resource(self, resource: Resource) -> None:
        """
        Releases a persistent resource. For a transient resource, this is a no-operation.

        :param resource: An instance of a resource to close.
        :type resource: Resource

        :return: Nothing. This call always succeeds.
        """

    @abstractmethod
    async def send_async(self, resource: Union[str, Resource], data: str, params: Optional[dict[str, str]] = None, timeout: float = 0.0) -> int:
        """
        Asynchronously send data to a resource. This function is intended to be used by behavioral models and platforms.
        The agents currently must use the synchronous option.

        :param resource: Either an instance of a resource, or its URL. When the URL is used, the resource is
            automatically created and destroyed within the call.
        :type resource: Union[str, Resource]

        :param data: Any data will be written to the resource. Technically, this can be empty and the contents can
            depend on the params.
        :type data: str

        :param params: Parameters for the sending operation. Such as choosing the write or append mode for the file
            resource.
        :type params: Optional[dict[str, str]]

        :param timeout: A maximum time given for the operation to finish. The semantic differs for simulated and
            emulated environments. For simulated ones, the timeout represents the actual length in virtual time units,
            whereas for emulated ones, the timeout is the maximum time to finish. That is, simulated send will always
            take timeout number of time units, and emulated will take up-to timeout time units.
        :type timeout: float

        :return: The number of bytes written. However, the sending operation is expected to write all of its contents.
            In case of some error an exception should be thrown.
        """

    @abstractmethod
    def send(self, resource: Union[str, Resource], data: str, params: Optional[dict[str, str]] = None, timeout: float = 0.0, callback_service: Optional[ActiveService] = None) -> None:
        """
        Synchronously send data to a resource. This function is expected to be used by agents, but it can be with a
        bit of over-engineering used by behavioral models and platforms as well. The main difference is that the
        results of the send operation are send inside a :class:`MessageType.RESOURCE` message.

        :param resource: Either an instance of a resource, or its URL. When the URL is used, the resource is
            automatically created and destroyed within the call.
        :type resource: Union[str, Resource]

        :param data: Any data will be written to the resource. Technically, this can be empty and the contents can
            depend on the params.
        :type data: str

        :param params: Parameters for the sending operation. Such as choosing the write or append mode for the file
            resource.
        :type params: Optional[dict[str, str]]

        :param timeout: A maximum time given for the operation to finish. The semantic differs for simulated and
            emulated environments. For simulated ones, the timeout represents the actual length in virtual time units,
            whereas for emulated ones, the timeout is the maximum time to finish. That is, simulated send will always
            take timeout number of time units, and emulated will take up-to timeout time units.
        :type timeout: float

        :param callback_service: A reference to an active service that will receive the result of the call within
            a resource message.
        :type callback_service: Optional[ActiveService] = None

        :return: None. The sending operation is always expected to write all of its contents. In case of some error
            an exception is thrown.
        """

    @abstractmethod
    async def fetch_async(self, resource: Union[str, Resource], params: Optional[dict[str, str]] = None, timeout: float = 0.0) -> Optional[str]:
        """
        Asynchronously reads data from a resource. This function is intended to be used by behavioral models and
        platforms. The agents currently must use the synchronous option.

        :param resource: Either an instance of a resource, or its URL. When the URL is used, the resource is
            automatically created and destroyed within the call.
        :type resource: Union[str, Resource]

        :param params: Parameters for the receiving operation.
        :type params: Optional[dict[str, str]]

        :param timeout: A maximum time given for the operation to finish. The semantic differs for simulated and
            emulated environments. For simulated ones, the timeout represents the actual length in virtual time units,
            whereas for emulated ones, the timeout is the maximum time to finish. That is, simulated send will always
            take timeout number of time units, and emulated will take up-to timeout time units.
        :type timeout: float

        :return: The data read from the resource. It may return none and still be correct, such as the HEAD HTML
            request.
        """

    @abstractmethod
    def fetch(self, resource: Union[str, Resource], params: Optional[dict[str, str]] = None, timeout: float = 0.0, callback_service: Optional[ActiveService] = None) -> None:
        """
        Synchronously reads data from a resource. This function is expected to be used by agents, but it can be with a
        bit of over-engineering used by behavioral models and platforms as well. The main difference is that the
        results of the fetch operation are send inside a :class:`MessageType.RESOURCE` message.

        :param resource: Either an instance of a resource, or its URL. When the URL is used, the resource is
            automatically created and destroyed within the call.
        :type resource: Union[str, Resource]

        :param params: Parameters for the receiving operation.
        :type params: Optional[dict[str, str]]

        :param timeout: A maximum time given for the operation to finish. The semantic differs for simulated and
            emulated environments. For simulated ones, the timeout represents the actual length in virtual time units,
            whereas for emulated ones, the timeout is the maximum time to finish. That is, simulated send will always
            take timeout number of time units, and emulated will take up-to timeout time units.
        :type timeout: float

        :param callback_service: A reference to an active service that will receive the result of the call within
            a resource message.
        :type callback_service: Optional[ActiveService] = None

        :return: The data read from the resource. It may return none and still be correct, such as the HEAD HTML
            request.
        """



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Tuple
from netaddr import IPNetwork, IPAddress
from uuid import uuid4

from cyst.api.environment.message import Message


@dataclass
class Route:
    """
    A route specifies which port should the traffic to specific network be routed through. Many routes can be specified
    for a node. If there is an overlap in network specification, than the resulting port is selected according to the
    route metrics (i.e., the lower the metric the higher the chance to be selected as the route). In case of a metric
    equality, the most specific network is selected.

    :param net: A network this route is related to.
    :type net: IPNetwork

    :param port: A port/interface index, where to route traffic to the particular network.
    :type port: int

    :param metric: A route metric used for deciding which route to use in case of network overlap.
    :type metric: int

    :param id: A unique identifier of the route. You very likely don't need to set it and just let it autogenerate.
    :type id: str
    """
    net: IPNetwork
    port: int
    metric: int = 100
    id: str = field(default_factory=lambda: str(uuid4()))

    # Custom comparison to enable sorting in a priority queue
    def __lt__(self, other: 'Route') -> bool:
        # Metric is a way to override the default longest-prefix routing
        if self.metric != other.metric:
            return self.metric < other.metric

        # This should usually suffice
        if self.net.prefixlen != other.net.prefixlen:
            # The comparison is inversed, because we want the longest prefix to have the lowest value and highest priority
            return self.net.prefixlen > other.net.prefixlen

        # This is just a fallback to have some stability in it
        return self.net.ip < other.net.ip


class Port(ABC):
    """
    A network port represents an abstraction of an ethernet port, with a given IP address and a given network. A network
    port does not support a default routing through a gateway and so it is used mostly for routers, which maintain
    their own routing tables based on the port indexes.
    """

    @property
    @abstractmethod
    def ip(self) -> Optional[IPAddress]:
        """
        Returns the IP address of the port.

        :rtype: Optional[IPAddress]
        """

    @property
    @abstractmethod
    def mask(self) -> Optional[str]:
        """
        Returns the network mask of the port.

        :rtype: Optional[str]
        """

    @property
    @abstractmethod
    def net(self) -> Optional[IPNetwork]:
        """
        Returns the network of the port.

        :rtype: Optional[IPNetwork]
        """

    @property
    @abstractmethod
    def connection(self) -> Optional['Connection']:
        """
        Returns the connection of the port.
        """


class Interface(Port, ABC):
    """
    A network interface represents an abstraction of an ethernet port, with a given IP address and a given network.
    The network interface automatically calculates the gateway and therefore enables a seamless networking.
    """

    @property
    @abstractmethod
    def gateway(self) -> Optional[IPAddress]:
        """
        Returns the IP address of the gateway.

        :rtype: Optional[IPAddress]
        """


class Connection(ABC):
    """
    Represents a connection between two network ports/interfaces. Connection supports setting of
    connection properties, such as delay or packet drops.
    """

    @property
    @abstractmethod
    def delay(self) -> int:
        """
        Returns the delay of the connection in simulated time units.

        :rtype: int
        """

    @property
    @abstractmethod
    def blocked(self) -> bool:
        """
        Returns whether the connection is blocked or not.

        :rtype: bool
        """

    @abstractmethod
    def set_params(self, delay: Optional[int] = None, blocked: Optional[bool] = None) -> None:
        """
        Sets the connection parameters, such as delay and blocked.

        :param delay: The delay of the connection in simulated time units.
        :type delay: int

        :param blocked: Whether the connection is blocked or not.
        :type blocked: bool
        """

    @abstractmethod
    def evaluate(self, message: Message) -> Tuple[int, Message]:
        """
        Evaluates the message based on the connection properties. The properties are checked in this order:

        1. The message is dropped: new response is formed from the message and delay is set to -1
        2. The message is delayed: the original message is returned alongside connections's delay
        3. The message is passed through: the original message is returned alongside 0 delay

        :param message: A message to evaluate.
        :type message: Message

        :return: A tuple of delay and message.
        """



import jsonpickle
import logging.config

from dataclasses import dataclass, field, fields, is_dataclass
from pathlib import Path
from typing import Dict, List, Union, Any, Callable, Optional, Type, Iterable
from uuid import uuid4

import netaddr

import cyst.core.environment.log
from cyst.api.configuration.infrastructure.log import LogType
from cyst.api.configuration.infrastructure.physical import PhysicalLocationConfig, PhysicalConnectionConfig, \
    PhysicalAccessConfig
from cyst.api.environment.configuration import GeneralConfiguration, ObjectType, ConfigurationObjectType
from cyst.api.environment.environment import Environment
from cyst.api.configuration.configuration import ConfigItem, ConfigParametrization, ConfigParameterSingle, \
    ConfigParameterGroup, ConfigParameterGroupType, ConfigParameter, ConfigParameterValueType
from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig
from cyst.api.configuration.infrastructure.infrastructure import InfrastructureConfig
from cyst.api.configuration.infrastructure.log import LogConfig, LogSource, log_defaults
from cyst.api.configuration.logic.access import AuthorizationConfig, AuthenticationProviderConfig, AccessSchemeConfig, \
    AuthorizationDomainConfig, FederatedAuthorizationConfig
from cyst.api.configuration.logic.data import DataConfig
from cyst.api.configuration.logic.exploit import VulnerableServiceConfig, ExploitParameterConfig, ExploitConfig
from cyst.api.configuration.network.elements import PortConfig, InterfaceConfig, ConnectionConfig, RouteConfig, SessionConfig
from cyst.api.configuration.network.firewall import FirewallChainConfig, FirewallConfig, FirewallPolicy, FirewallChainType
from cyst.api.configuration.network.network import NetworkConfig
from cyst.api.configuration.network.router import RouterConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.utils.counter import Counter


class Configurator:
    def __init__(self, environment: Environment):
        self._env = environment
        self._refs: Dict[str, Any] = {}
        self._connections: List[ConnectionConfig] = []
        self._sessions: List[SessionConfig] = []
        self._nodes: List[NodeConfig] = []
        self._routers: List[RouterConfig] = []
        self._active_services: List[ActiveServiceConfig] = []
        self._passive_services: List[PassiveServiceConfig] = []
        self._firewalls: List[FirewallConfig] = []
        self._interfaces: List[Union[InterfaceConfig, PortConfig]] = []
        self._authorizations: List[AuthorizationConfig] = []
        self._data: List[DataConfig] = []
        self._exploits: List[ExploitConfig] = []
        self._authentication_providers: List[AuthenticationProviderConfig] = []
        self._access_schemes: List[AccessSchemeConfig] = []
        self._authorization_domains: List[AuthorizationDomainConfig] = []
        self._logs: List[LogConfig] = []
        self._physical_access_config: List[PhysicalAccessConfig] = []
        self._physical_location_config: List[PhysicalLocationConfig] = []
        self._physical_connection_config: List[PhysicalConnectionConfig] = []
        self._config_parametrization: ConfigParametrization | None = None

    def __getstate__(self):
        result = self.__dict__
        return result

    def __setstate__(self, state):
        self.__dict__.update(state)

    def reset(self):
        self._refs.clear()
        self._connections.clear()
        self._sessions.clear()
        self._nodes.clear()
        self._routers.clear()
        self._active_services.clear()
        self._passive_services.clear()
        self._firewalls.clear()
        self._interfaces.clear()
        self._authorizations.clear()
        self._data.clear()
        self._exploits.clear()
        self._authentication_providers.clear()
        self._access_schemes.clear()
        self._authorization_domains.clear()
        self._logs.clear()

    # ------------------------------------------------------------------------------------------------------------------
    # All these _process_XXX functions resolve nested members to their id. In the end of the preprocessing, there should
    # be only a flat configuration with ids and no nesting (see members above)
    # ------------------------------------------------------------------------------------------------------------------
    def _process_NetworkConfig(self, cfg: NetworkConfig) -> NetworkConfig:
        node_refs = []
        conn_refs = []

        for node in cfg.nodes:
            if isinstance(node, str):
                node_refs.append(node)
            else:
                node_refs.append(self._process_cfg_item(node))

        for conn in cfg.connections:
            if isinstance(conn, str):
                conn_refs.append(conn)
            else:
                conn_refs.append(self._process_cfg_item(conn))

        cfg.nodes = node_refs
        cfg.connections = conn_refs

        return cfg

    def _process_ConnectionConfig(self, cfg: ConnectionConfig) -> str:
        self._connections.append(cfg)
        self._refs[cfg.ref] = cfg
        return cfg.ref

    def _process_SessionConfig(self, cfg: SessionConfig) -> str:
        self._sessions.append(cfg)
        self._refs[cfg.ref] = cfg
        return cfg.ref

    def _process_RouterConfig(self, cfg: RouterConfig):
        interface_refs = []
        traffic_processor_refs = []

        for interface in cfg.interfaces:
            if isinstance(interface, str):
                interface_refs.append(interface)
            else:
                interface_refs.append(self._process_cfg_item(interface))

        for processor in cfg.traffic_processors:
            if isinstance(processor, str):
                traffic_processor_refs.append(processor)
            else:
                traffic_processor_refs.append(self._process_cfg_item(processor))

        for route in cfg.routing_table:
            self._process_RouteConfig(route)

        cfg.interfaces = interface_refs
        cfg.traffic_processors = traffic_processor_refs

        self._routers.append(cfg)
        self._refs[cfg.ref] = cfg
        return cfg.ref

    def _process_NodeConfig(self, cfg: NodeConfig) -> str:
        passive_service_refs = []
        active_service_refs = []
        traffic_processor_refs = []
        interface_refs = []

        for service in cfg.passive_services:
            if isinstance(service, str):
                passive_service_refs.append(service)
            else:
                passive_service_refs.append(self._process_cfg_item(service))

        for service in cfg.active_services:
            if isinstance(service, str):
                active_service_refs.append(service)
            else:
                active_service_refs.append(self._process_cfg_item(service))

        for processor in cfg.traffic_processors:
            if isinstance(processor, str):
                traffic_processor_refs.append(processor)
            else:
                traffic_processor_refs.append(self._process_cfg_item(processor))

        for interface in cfg.interfaces:
            if isinstance(interface, str):
                interface_refs.append(interface)
            else:
                interface_refs.append(self._process_cfg_item(interface))

        cfg.passive_services = passive_service_refs
        cfg.active_services = active_service_refs
        cfg.traffic_processors = traffic_processor_refs
        cfg.interfaces = interface_refs

        self._nodes.append(cfg)
        self._refs[cfg.ref] = cfg

        return cfg.ref

    def _process_PortConfig(self, cfg: PortConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._interfaces.append(cfg)
        return cfg.ref

    def _process_InterfaceConfig(self, cfg: InterfaceConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._interfaces.append(cfg)
        return cfg.ref

    def _process_ActiveServiceConfig(self, cfg: ActiveServiceConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._active_services.append(cfg)
        return cfg.ref

    def _process_FirewallConfig(self, cfg: FirewallConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._firewalls.append(cfg)
        return cfg.ref

    def _process_PassiveServiceConfig(self, cfg: PassiveServiceConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._passive_services.append(cfg)

        public_data_refs = []
        private_data_refs = []
        public_auth_refs = []
        private_auth_refs = []

        for data in cfg.public_data:
            if isinstance(data, str):
                public_data_refs.append(data)
            else:
                public_data_refs.append(self._process_cfg_item(data))

        for data in cfg.private_data:
            if isinstance(data, str):
                private_data_refs.append(data)
            else:
                private_data_refs.append(self._process_cfg_item(data))

        for auth in cfg.public_authorizations:
            if isinstance(auth, str):
                public_auth_refs.append(auth)
            else:
                public_auth_refs.append(self._process_cfg_item(auth))

        for auth in cfg.private_authorizations:
            if isinstance(auth, str):
                private_auth_refs.append(auth)
            else:
                private_auth_refs.append(self._process_cfg_item(auth))

        cfg.public_data = public_data_refs
        cfg.private_data = private_data_refs
        cfg.public_authorizations = public_auth_refs
        cfg.private_authorizations = private_auth_refs

        auth_provider_refs = []
        for provider in cfg.authentication_providers:
            if isinstance(provider, str):
                auth_provider_refs.append(provider)
            else:
                auth_provider_refs.append(self._process_cfg_item(provider))
        cfg.authentication_providers = auth_provider_refs

        access_scheme_refs = []
        for scheme in cfg.access_schemes:
            if isinstance(scheme, str):
                access_scheme_refs.append(scheme)
            else:
                access_scheme_refs.append(self._process_cfg_item(scheme))
        cfg.access_schemes = access_scheme_refs

        return cfg.ref

    def _process_AuthorizationConfig(self, cfg: AuthorizationConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._authorizations.append(cfg)
        return cfg.ref

    def _process_DataConfig(self, cfg: DataConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._data.append(cfg)
        return cfg.ref

    def _process_ExploitConfig(self, cfg: ExploitConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._exploits.append(cfg)
        return cfg.ref

    def _process_AuthenticationProviderConfig(self, cfg: AuthenticationProviderConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._authentication_providers.append(cfg)
        return cfg.ref

    def _process_AccessSchemeConfig(self, cfg: AccessSchemeConfig) -> str:

        auth_provider_refs = []
        for provider in cfg.authentication_providers:
            if isinstance(provider, str):
                auth_provider_refs.append(provider)
            else:
                auth_provider_refs.append(self._process_cfg_item(provider))
        cfg.authentication_providers = auth_provider_refs

        if not isinstance(cfg.authorization_domain, str):
            cfg.authorization_domain = self._process_cfg_item(cfg.authorization_domain)

        self._refs[cfg.ref] = cfg
        self._access_schemes.append(cfg)
        return cfg.ref

    def _process_AuthorizationDomainConfig(self, cfg: AuthorizationDomainConfig) -> str:

        authorization_refs = []
        for auth in cfg.authorizations:
            if isinstance(auth, str):
                authorization_refs.append(auth)
            else:
                authorization_refs.append(self._process_cfg_item(auth))
        cfg.authorizations = authorization_refs

        self._refs[cfg.ref] = cfg
        self._authorization_domains.append(cfg)
        return cfg.ref

    def _process_FederatedAuthorizationConfig(self, cfg: FederatedAuthorizationConfig):

        # TODO : Ask how this is meant to be handled

        self._refs[cfg.ref] = cfg
        return cfg.ref

    def _process_RouteConfig(self, cfg: RouteConfig) -> str:

        self._refs[cfg.ref] = cfg
        return cfg.ref

    def _process_LogConfig(self, cfg: LogConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._logs.append(cfg)
        return cfg.ref

    def _process_InfrastructureConfig(self, cfg: InfrastructureConfig) -> str:
        self._refs[cfg.ref] = cfg
        for log in cfg.log:
            self._process_LogConfig(log)
        return cfg.ref

    def _process_PhysicalAccessConfig(self, cfg: PhysicalAccessConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._physical_access_config.append(cfg)
        return cfg.ref

    def _process_PhysicalLocationConfig(self, cfg: PhysicalLocationConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._physical_location_config.append(cfg)
        for access in cfg.access:
            self._process_PhysicalAccessConfig(access)
        return cfg.ref

    def _process_PhysicalConnectionConfig(self, cfg: PhysicalConnectionConfig) -> str:
        self._refs[cfg.ref] = cfg
        self._physical_connection_config.append(cfg)
        return cfg.ref

    def _process_ConfigParametrization(self, cfg: ConfigParametrization) -> str:
        self._refs[cfg.ref] = cfg
        self._config_parametrization = cfg
        return cfg.ref

    def _process_default(self, cfg):
        raise ValueError(f"Unknown config type provided: {type(cfg)}")

    def _process_cfg_item(self, cfg: Any) -> str:
        if hasattr(cfg, "ref") and cfg.ref in self._refs:
            if self._refs[cfg.ref] != cfg:
                raise ValueError("Duplicate identifier for different configuration objects found: {}".format(cfg.ref))
            else:
                return cfg.ref
        else:
            fn: Callable[[ConfigItem], str] = getattr(self, "_process_" + type(cfg).__name__, self._process_default)
            return fn(cfg)

    # ------------------------------------------------------------------------------------------------------------------
    # Parametrization
    def parse_parameters(self, parametrization_config: ConfigParametrization, parameters: dict[str, Any]):
        """
        Parse parameters into the final form for easier parametrization and fill out the right type of value.
        """
        parsed_parameters = {}

        # Fill group parameters
        for parameter in parametrization_config.parameters:
            if isinstance(parameter, ConfigParameterGroup):
                for group_entry in parameter.options:
                    if group_entry.parameter_id in parameters[parameter.parameter_id]:
                        parsed_parameters[group_entry.parameter_id] = self.get_config_parameter_value(group_entry.value, parameter.value_type)
            elif isinstance(parameter, ConfigParameterSingle):
                parameter_value = parameters[parameter.parameter_id]
                parsed_parameters[parameter.parameter_id] = self.get_config_parameter_value(parameter_value,
                                                                                              parameter.value_type)

        return parsed_parameters


    def check_and_default_parameters(self, parametrization_config: ConfigParametrization, input_parameters: dict[str, Any] | None):
        """
        Check the correctness of group parameters and fill in the defaults if the parameters are missing.
        """
        # Initialize parameters as an empty dictionary if it is None
        if input_parameters is None:
            input_parameters = {}

        def validate_group_entries(group_param, frontend_value):
            if group_param.group_type == ConfigParameterGroupType.ONE:
                if len(frontend_value) != 1:
                    raise ValueError(f"Group '{group_param.parameter_id}' must have exactly one value")
            elif group_param.group_type == ConfigParameterGroupType.ANY:
                if len(frontend_value) < 1:
                    raise ValueError(f"Group '{group_param.parameter_id}' must have at least one value")


        for parameter in parametrization_config.parameters:
            if isinstance(parameter, ConfigParameterSingle):
                if parameter.parameter_id not in input_parameters:
                    input_parameters[parameter.parameter_id] = parameter.default
            elif isinstance(parameter, ConfigParameterGroup):
                if parameter.parameter_id not in input_parameters:
                    input_parameters[parameter.parameter_id] = parameter.default
                validate_group_entries(parameter, input_parameters[parameter.parameter_id])
        return input_parameters

    def get_config_parameter_value(self, value: str, value_type: ConfigParameterValueType) -> str | None:
        """
        Get the value of a configuration parameter based on its type.
        """
        match value_type:
            case ConfigParameterValueType.VALUE:
                return value
            case ConfigParameterValueType.REF:
                if value not in self._refs:
                    raise ValueError(f"Reference: '{value}' not found in configuration items")
                return value
            case ConfigParameterValueType.REF_COPY:
                try:
                    config_item_copy = self._refs[value]()
                    self._process_cfg_item(config_item_copy)
                    return config_item_copy.ref
                except KeyError:
                    raise ValueError(f"Reference: '{value}' not found in configuration items")

    def combine_config_items_with_parameters(self, parameters: dict[str, Any] | None, config_items: Iterable[ConfigItem]):
        """
        Fill in the correct parameter values in place of ConfigParameter instances in the configuration items.
        """
        # check the correctness of group parameters and fill in the defaults if the parameters are missing
        parameters = self.check_and_default_parameters(self._config_parametrization, parameters)

        # parse parameters into the final form for easier parametrization and fill out the right type of value.
        parsed_parameters = self.parse_parameters(self._config_parametrization, parameters)

        def set_config_parameters(obj):

            if isinstance(obj, ConfigParameter):
                # If it's a ConfigParameter, replace it with the value from parameters if available
                if (parameter_value := parsed_parameters.get(obj.id)) is not None:
                    return parameter_value
                else:
                    return obj
            elif isinstance(obj, dict):
                # Recursively handle dictionary items
                modified_dict = {key: set_config_parameters(value) for key, value in obj.items()}
                # Remove ConfigParameter instances from the dictionary
                return {k: v for k, v in modified_dict.items() if not isinstance(v, ConfigParameter)}
            elif isinstance(obj, list):
                # Recursively handle list elements
                modified_list = [set_config_parameters(item) for item in obj]
                # Remove ConfigParameter instances from the list
                return [item for item in modified_list if not isinstance(item, ConfigParameter)]
            elif is_dataclass(obj):
                # Handle dataclass objects by iterating through their fields
                for field in fields(obj):
                    new_value = set_config_parameters(getattr(obj, field.name))
                    if isinstance(new_value, ConfigParameter):
                        raise ValueError(f"Parameter id of '{new_value.id}' is missing in parametrization configuration")
                    setattr(obj, field.name, set_config_parameters(getattr(obj, field.name)))
            return obj

        # Apply the set_config_parameters function to each item in the configuration
        for item in config_items:
            set_config_parameters(item)
    # --------------------------------------------------------------------------------------------------------------

    def preprocess(self,
                   parameters: dict[str, Any],
                  *configs: Union[NetworkConfig, ConnectionConfig, RouterConfig, NodeConfig,
                                  InterfaceConfig, ActiveServiceConfig, PassiveServiceConfig,
                                  AuthorizationConfig, DataConfig, PhysicalAccessConfig, PhysicalLocationConfig,
                                  PhysicalConnectionConfig]) -> 'Configurator':

        # --------------------------------------------------------------------------------------------------------------
        # Process all provided items and do a complete id->cfg mapping
        if type(configs) == tuple and len(configs) == 1 and type(configs[0]) == list:
            configs = configs[0]
        for cfg in configs:
            self._process_cfg_item(cfg)

        if self._config_parametrization:
            self.combine_config_items_with_parameters(parameters, self._refs.values())

        # Create interface configurations for connections with one -1 port id
        for connection_cfg in self._connections:
            if connection_cfg.dst_port == -1 or connection_cfg.src_port == -1:
                if connection_cfg.dst_port == -1:
                    target_ref = connection_cfg.dst_ref
                    source_ref = connection_cfg.src_ref
                    source_port = connection_cfg.src_port
                else:
                    target_ref = connection_cfg.src_ref
                    source_ref = connection_cfg.dst_ref
                    source_port = connection_cfg.dst_port

                target_config = self._refs[target_ref]
                source_config = self._refs[source_ref]

                new_interface_index = len(target_config.interfaces)
                new_interface_ref = str(uuid4())

                source_net = self._refs[source_config.interfaces[source_port]].net

                # The port is on the leaf, use the first free address from the router
                if isinstance(target_config, NodeConfig):
                    allocated_ips = []

                    # The following is not pretty and not optimal, however...
                    # Get all nodes connected to the same router we are trying to connect to now
                    connected_refs = []
                    for conn in self._connections:
                        if source_config.ref == conn.src_ref:
                            connected_refs.append(conn.dst_ref)
                        elif source_config.ref == conn.dst_ref:
                            connected_refs.append(conn.src_ref)

                    for node_cfg in self._nodes:
                        if node_cfg.ref in connected_refs:
                            for iface_ref in node_cfg.interfaces:
                                iface_cfg = self._refs[iface_ref]
                                allocated_ips.append(iface_cfg.ip)

                    # Add router addresses
                    for iface in source_config.interfaces:
                        allocated_ips.append(self._refs[iface].ip)

                    new_ip = None
                    for ip in source_net.iter_hosts():
                        if ip not in allocated_ips:
                            new_ip = ip
                            break

                    if not new_ip:
                        raise RuntimeError(f"Cannot find a free IP for target in a connection {connection_cfg}.")

                    iface = InterfaceConfig(
                        ip=new_ip,
                        net=source_net,
                        index=new_interface_index,
                        ref=new_interface_ref
                    )

                # The port is on the router, just copy the configuration from the leaf node
                elif isinstance(target_config, RouterConfig):
                    # Set the IP to the first address in the net
                    iface = InterfaceConfig(
                        ip=netaddr.IPAddress(source_net.first + 1),
                        net=source_net,
                        index=new_interface_index,
                        ref=new_interface_ref
                    )

                else:
                    raise RuntimeError(f"Attempting to connect something else than a node or router: {target_config}.")

                self._interfaces.append(iface)
                self._refs[new_interface_ref] = iface
                target_config.interfaces.append(new_interface_ref)

                if connection_cfg.dst_port == -1:
                    connection_cfg.dst_port = new_interface_index
                else:
                    connection_cfg.src_port = new_interface_index

        # Construct names and IDs
        processed_ref_set = set()

        # TODO: I am ignoring the whole network configuration, because nobody is using it. Should it be removed?
        def name_config_item(parent_id: str, item: ConfigItem):
            # Some configuration items are shared in multiple places, but for each ref there should be only one object
            # with distinct id. So if we hit an already used ref, we assign it an existing id.
            if item.ref in processed_ref_set:
                other = self._refs[item.ref]
                item.name = other.name
                item.id = other.id
            else:
                use_counter = False
                if not item.name:
                    # This is for the case when someone directly sets a name to empty string
                    if item.id:
                        item.name = item.id
                    else:
                        item.name = str(uuid4())
                elif item.name.startswith("__"):
                    item.name = item.name.removeprefix("__")
                    use_counter = True

                if not item.id:
                    partial_id = parent_id + "." + item.name if parent_id else item.name

                    if use_counter:
                        item_number = str(Counter().get(partial_id))
                        item.name = item.name + "_" + item_number
                        item.id = partial_id + "_" + item_number
                    else:
                        item.id = partial_id

                processed_ref_set.add(item.ref)

            for key, value in item.__dict__.items():
                if isinstance(value, ConfigItem):
                    name_config_item(item.id, value)
                elif isinstance(value, list):
                    for x in value:
                        if isinstance(x, ConfigItem):
                            name_config_item(item.id, x)

        config_tree = self.get_configuration()
        for item in config_tree:
            name_config_item("", item)

        return self

    # ------------------------------------------------------------------------------------------------------------------
    # Configuration of global stuff that is not relegated to platforms
    def configure(self):

        # --------------------------------------------------------------------------------------------------------------
        # Logs:
        # Get defaults
        log_configs = {}
        for log in log_defaults:
            log_configs[log.source] = log

        # Overwrite with user configuration
        for cfg in self._logs:
            log_configs[cfg.source] = cfg

        default_fields = {
            "virtual_time": self._env.resources.clock.current_time,
            "run_id": self._env.infrastructure.statistics.run_id
        }

        formatters = {
            '__text': {
                'format': "Real time: [%(asctime)s], Virtual time: [%(virtual_time).2f] :: %(name)s — %(levelname)s :: %(message)s",
                'class': 'cyst.core.environment.log.CYSTFormatter',
                'defaults': default_fields
            },
            '__json': {
                'format': '{"run_id": "%(run_id)s", "log_time": "%(asctime)s", "log_level": "%(levelname)s", '
                          '"virtual_time": "%(virtual_time).2f", "component": "%(name)s", "message": "%(message)s"}',
                'class': 'cyst.core.environment.log.CYSTFormatter',
                'defaults': default_fields
            },
        }

        handlers = {}
        loggers = {}

        log_source_map = {
            LogSource.SYSTEM: 'system',
            LogSource.MESSAGING: 'messaging',
            LogSource.MODEL: 'model',
            LogSource.SERVICE: 'service'
        }

        for cfg in log_configs.values():
            if not cfg.log_console and not cfg.log_file:
                continue

            handler_console = {}
            handler_file = {}

            if cfg.log_console:
                handler_console = {
                    'class': 'logging.StreamHandler',
                    'formatter': '__text' if cfg.log_type == LogType.TEXT else '__json',
                    'level': cfg.log_level,
                    'stream': 'ext://sys.stdout'
                }

            if cfg.log_file and cfg.file_path:
                if self._env.infrastructure.runtime_configuration.run_id_log_suffix:
                    p = Path(cfg.file_path)
                    path = p.parent / Path(f"{p.stem}-{self._env.infrastructure.runtime_configuration.run_id}{p.suffix}")
                else:
                    path = Path(cfg.file_path)

                path.parent.mkdir(parents=True, exist_ok=True)

                handler_file = {
                    'class': 'logging.FileHandler',
                    'formatter': '__text' if cfg.log_type == LogType.TEXT else '__json',
                    'level': cfg.log_level,
                    'filename': path
                }


            if handler_console or handler_file:
                handler_list = []
                if handler_console:
                    id = "__console__" + log_source_map[cfg.source]
                    handlers[id] = handler_console
                    handler_list.append(id)
                if handler_file:
                    id = "__file__" + log_source_map[cfg.source]
                    handlers[id] = handler_file
                    handler_list.append(id)

                loggers[log_source_map[cfg.source]] = {
                    'handlers': handler_list,
                    'level': cfg.log_level
                }

        log_config_dict = {
            'version': 1,
            'formatters': formatters,
            'handlers': handlers,
            'loggers': loggers
        }

        logging.config.dictConfig(log_config_dict)

        for location_cfg in self._physical_location_config:
            self._env.configuration.physical.create_physical_location(location_id=location_cfg.id)
            for physical_access_cfg in location_cfg.access:
                physical_access = self._env.configuration.physical.create_physical_access(
                    identity=physical_access_cfg.identity,
                    time_from=physical_access_cfg.time_from,
                    time_to=physical_access_cfg.time_to
                )
                self._env.configuration.physical.add_physical_access(location_cfg.id, physical_access)

            for asset in location_cfg.assets:
                self._env.configuration.physical.place_asset(location_cfg.id, asset)

        for physical_connection_cfg in self._physical_connection_config:
            self._env.configuration.physical.add_physical_connection(
                origin=physical_connection_cfg.origin,
                destination=physical_connection_cfg.destination,
                travel_time=physical_connection_cfg.travel_time
            )

    def get_configuration_by_id(self, id: str) -> Optional[Any]:
        if id not in self._refs:
            return None
        else:
            return self._refs[id]

    def _resolve_config_item(self, item: ConfigItem) -> ConfigItem:
        replaced = {}
        for key, value in item.__dict__.items():
            if isinstance(value, str) and not key.endswith("ref") and value in self._refs:
                replaced[key] = self._resolve_config_item(self._refs[value])
            if isinstance(value, list):
                tmp = []
                for element in value:
                    if isinstance(element, str) and element in self._refs:
                        tmp.append(self._resolve_config_item(self._refs[element]))
                if tmp:
                    replaced[key] = tmp
        item.__dict__.update(replaced)
        return item

    def get_configuration(self) -> List[ConfigItem]:
        top_level = [*self._nodes, *self._routers, *self._connections, *self._exploits, *self._sessions]
        result = []
        for item in top_level:
            result.append(self._resolve_config_item(item))
        return result


# ----------------------------------------------------------------------------------------------------------------------
class GeneralConfigurationImpl(GeneralConfiguration):

    def __init__(self, env: Environment) -> None:
        self._configurator = Configurator(env)
        self._env = env

    def preprocess(self, parameters: dict[str, Any], *config) -> None:
        self._configurator.preprocess(parameters, *config)

    def configure(self) -> None:
        self._configurator.configure()

    def get_configuration(self) -> List[ConfigItem]:
        return self._configurator.get_configuration()

    def save_configuration(self, indent: Optional[int]) -> str:
        return jsonpickle.encode(self._configurator.get_configuration(), make_refs=False, indent=indent)

    def load_configuration(self, config: str) -> List[ConfigItem]:
        return jsonpickle.decode(config)

    def get_configuration_by_id(self, id: str,
                                configuration_type: Type[ConfigurationObjectType]) -> ConfigurationObjectType:

        c = self._configurator.get_configuration_by_id(id)
        if not isinstance(c, configuration_type):
            raise AttributeError(
                "Attempting to cast configuration object with id: {} to an incompatible type: {}".format(id,
                                                                                                         str(configuration_type)))
        return c

    def get_object_by_id(self, id: str, object_type: Type[ObjectType]) -> ObjectType:
        if self._env.platform:
            return self._env.platform.configuration.general.get_object_by_id(id, object_type)

    @staticmethod
    def cast_from(o: GeneralConfiguration) -> 'GeneralConfigurationImpl':
        if isinstance(o, GeneralConfigurationImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the GeneralConfiguration interface")



from abc import ABC, abstractmethod
from enum import Enum
from semver import VersionInfo
from typing import Dict


class ExploitCategory(Enum):
    """
    Exploit category represents the high-level taxonomy of exploit effects.

    Possible values:
        :NONE: Either the category was not selected or is not one of the other possible values. This category should
            be used sparsely and after due consideration.
        :AUTH_MANIPULATION: The exploit is used to affect authentications or authorizations, such as privilege
            escalation or user manipulation.
        :DATA_MANIPULATION: The exploit is used to affect data, mostly through giving access to said data.
        :CODE_EXECUTION: The exploit enables execution of arbitrary code in the context of the action, so the result
            can be an opened session, new malware instance, etc.
    """
    NONE = 0,
    AUTH_MANIPULATION = 1,
    DATA_MANIPULATION = 2,
    CODE_EXECUTION = 4


class ExploitLocality(Enum):
    """
    The area of effect of the exploit, i.e., whether it can be used remotely, or whether only locally. Note that having
    a session opened to a remote machine counts as having a local access to the machine.

    Possible values:
        :NONE: Misconfiguration sentinel. No exploit should have this locality.
        :LOCAL: Exploit usable only locally.
        :REMOTE: Exploit usable over network.
    """
    NONE = 0
    LOCAL = 1
    REMOTE = 2


class ExploitParameterType(Enum):
    """
    Exploits support further parametrization. The parametrization works in a key-value manner, but unlike with actions,
    there is no domain specification of values, so they are all strings. This will likely change in the future. But
    at least keys (aka. parameter types have an explicit domain).

    Some parameters are mutable, i.e., able to be changed by the users, and some are immutable, which mostly describe
    in more details the effect of the exploit.

    Possible mutable values:
        :NONE: Misconfiguration sentinel. No exploit parameter should have this type.
        :IDENTITY: The parameter should contain a name of a user within a simulation.
        :ADDRESS: The parameter should contain an IP address.
        :SERVICE: The parameter should contain a service name.

    Possible immutable values:
        :IMPACT_IDENTITY: A successful exploitation will provide authorization for a selected few identities, or for
            all. Values: (ENUMERATED, ALL).
        :IMPACT_NODE: A successful exploitation will provide authorization for a selected few nodes, or for all.
            Values: (ENUMERATED, ALL).
        :IMPACT_SERVICE: A successful exploitation will provide authorization for a selected few services, or for all.
            Values: (ENUMERATED, ALL).
        :ENABLE_ELEVATED_ACCESS: An exploit enables to get root access. Values: (TRUE, FALSE).

    """
    NONE = 0,
    IDENTITY = 1,
    ADDRESS = 2,
    SERVICE = 3,
    IMPACT_IDENTITY = 1001,
    IMPACT_NODE = 1002,
    IMPACT_SERVICE = 1003,
    ENABLE_ELEVATED_ACCESS = 1004


class ExploitParameter(ABC):
    """
    A representation of exploit parameter.

    :param exploit_type: A type of the exploit parameter. TODO: Rename to include parameter.
    :type exploit_type: ExploitParameterType

    :param value: A value of the exploit parameter.
    :type value: str

    :param immutable: An immutability of the exploit parameter.
    :type immutable: bool
    """
    @abstractmethod
    def __init__(self, exploit_type: ExploitParameterType, value: str = "", immutable: bool = False) -> None:
        """ Initialization method for ExploitParameter """

    @property
    @abstractmethod
    def type(self) -> ExploitParameterType:
        """
        Returns a type of the exploit parameter.

        :rtype: ExploitParameterType
        """

    @property
    @abstractmethod
    def value(self) -> str:
        """
        Returns a value of the exploit parameter.

        :rtype: str
        """

    @abstractmethod
    def set_value(self, val: str) -> None:
        """
        Sets the value of the parameter, provided it is mutable one.

        :param val: A new value.
        :type val: str
        """


class VulnerableService(ABC):
    """
    A specification of a service and its version that are vulnerable for the purpose of exploit description.
    """

    @property
    @abstractmethod
    def id(self) -> str:
        """
        Returns the name of the service.

        :rtype: str
        """

    @property
    @abstractmethod
    def min_version(self) -> VersionInfo:
        """
        Returns the minimal affected version.

        :rtype: VersionInfo
        """

    @property
    @abstractmethod
    def max_version(self) -> VersionInfo:
        """
        Returns the maximal affected version.

        :rtype: VersionInfo
        """


class Exploit(ABC):
    """
    A description of a working exploit.
    """

    @property
    @abstractmethod
    def id(self) -> str:
        """
        Returns the unique identification of the exploit.

        :rtype: str
        """

    @property
    @abstractmethod
    def services(self) -> Dict[str, VulnerableService]:
        """
        Returns the collection of services that are affected by this exploit.

        :rtype: Dict[str, VulnerableService]:
        """

    @property
    @abstractmethod
    def locality(self) -> ExploitLocality:
        """
        Returns the locality of the exploit.

        :rtype: ExploitLocality
        """

    @property
    @abstractmethod
    def category(self) -> ExploitCategory:
        """
        Returns the category of the exploit.

        :rtype: ExploitCategory
        """

    @property
    @abstractmethod
    def parameters(self) -> Dict[ExploitParameterType, ExploitParameter]:
        """
        Returns the collection of exploit parameters.

        :rtype: Dict[ExploitParameterType, ExploitParameter]
        """



from abc import ABC, abstractmethod
from dataclasses import dataclass
from deprecated.sphinx import versionchanged, versionadded
from typing import List, Optional, Tuple, Union, Dict, Any, Callable

from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.message import Message, Signal
from cyst.api.environment.stats import Statistics
from cyst.api.host.service import ActiveService
from cyst.api.logic.access import AccessLevel
from cyst.api.logic.action import Action, ActionDescription
from cyst.api.logic.exploit import Exploit, ExploitCategory
from cyst.api.network.node import Node


class ActionStore(ABC):
    """
    Action store provides access to actions that are available to services.
    """

    @abstractmethod
    def get(self, id: str = "") -> Optional[Action]:
        """
        Returns an action with given ID. This function makes a copy of the object, which is present in the store. This
        is a preferred variant, because any parameters set on that action would propagate to the store.

        :param id: A unique ID of the action.
        :type id: str

        :return: An action, if there is one with such ID and for such execution environment.
        """

    @abstractmethod
    def get_ref(self, id: str = "") -> Optional[Action]:
        """
        Return an action with give ID. This function returns a reference to the object stored in the store and any
        parameter alterations will propagate to all subsequent queries for this action.

        :param id: A unique ID of the action.
        :type id: str

        :return: An action, if there is one with such ID and for such execution environment.
        """

    @abstractmethod
    def get_prefixed(self, prefix: str = "") -> List[Action]:
        """
        Gets a list of actions, whose ID starts with a given string. This is usually done to get access to the entire
        namespace of a particular behavioral model.

        The list will contain copies of actions present in the store. Getting multiple references in one call is not
        supported.

        :param prefix: The prefix all actions IDs must share.
        :type prefix: str

        :return: A list of actions with the same prefix.
        """

    @abstractmethod
    def add(self, action: ActionDescription) -> None:
        """
        Adds a new action to the store. This function should be used in two cases:

        * Adding new action for a behavioral model. Such action must have a processing function mapped to the
          action ID.

        * Adding new action for intra-agent communication. There is no requirement on the action form, however
          an exception will be thrown, if this action is directed to a passive service, as the system will have
          no idea how to process it.

        :param action: A description of the action to add.
        :type action: ActionDescription

        :return: None
        """


class ExploitStore(ABC):
    """
    Exploit store provides access to exploits that can be used together with actions. Unlike the action store,
    runtime definition of exploits by services is not permitted. This must be done through the
    :class:`cyst.api.environment.configuration.ExploitConfiguration` interface.
    """

    @abstractmethod
    def get_exploit(self, id: str = "", service: str = "", category: ExploitCategory = ExploitCategory.NONE) -> Optional[List[Exploit]]:
        """
        Gets an exploit, which satisfy all the parameters.

        :param id: An explicit ID of an exploit.
        :type id: str

        :param service: An ID of a service the exploit can be used at.
        :type service: str

        :param category: A category that the exploit should have. If the ExploitCategory.NONE is set, then the category
            is not considered when retrieving the exploits.
        :type category: ExploitCategory

        :return: A list of exploits satisfying the parameters.
        """

    @abstractmethod
    def evaluate_exploit(self, exploit: Union[str, Exploit], message: Message, node: Node) -> Tuple[bool, str]:
        """
        Evaluates, whether the provided exploit is applicable, given the message which carries the relevant action and
        a concrete node. TODO: This interface is cumbersome. While this is best fit for the data that interpreters
        receive, it is confusing at best.

        :param exploit: The ID of the exploit or its instance.
        :type exploit: Union[str, Exploit]

        :param message: An instance of the message which carried the exploit.
        :type message: Message

        :param node: An instance of the node, where the exploit is being applied.
        :type node: Node

        :return: (True, _) if exploit is applicable, (False, reason) otherwise.
        """


@versionadded(version="0.6.0")
class ServiceStore(ABC):
    """
    Service store provides a unified interface for creating active services. Due to centrality of this concept to all
    CYST, regardless of the platform it uses, all services must be instantiated through this store.
    """

    @abstractmethod
    def create_active_service(self, type: str, owner: str, name: str, node: Node,
                              service_access_level: AccessLevel = AccessLevel.LIMITED,
                              configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[ActiveService]:
        """
        Creates an active service instance.

        :param type: The name of the active service, under which it is registered into the system.
        :param owner: The identity of user, under whose identity this service should be running.
        :param name: The name of the service, under which it is present at the node. Currently, it is used instead
            of ports to route messages. TODO: The name/id system is retarded and has to be changed.
        :param node: The node at which the service should be instantiated.
        :param service_access_level: The level of access the service has on the system.
        :param id: System-wide unique ID of the service. Unless overridden, it will have the form of
            `<node_id>.<service_name>`.
        :param configuration: A dictionary with arbitrary configuration items.

        :return: The instantiated active service or None if there was an error.
        """

    @abstractmethod
    def get_active_service(self, id) -> Optional[ActiveService]:
        """
        Returns an already instantiated active service, given its full id (that is, not just a name).

        :param id:
        :return:
        """


@versionadded(version="0.6.0")
class DataStore(ABC):

    @abstractmethod
    def add_action(self, action: ActionModel) -> None:
        """
        Store information about a resolved action, i.e., after completing the request-response cycle.

        :param action: An action description

        :return: None
        """

    @abstractmethod
    def add_message(self, message: Message) -> None:
        """
        Store information about a message. While in general, the message information is available only at the point of
        dispatching the message, platforms that have more control over the message passing process (e.g., simulation
        platform of CYST) may report the same message multiple times during the message passing. Additional information
        shall then be passed through the .platform_specific attributed, which gets erased before entering user-facing
        side of the code.

        :param message: The message to store

        :return: None
        """

    @abstractmethod
    def add_statistics(self, statistics: Statistics) -> None:
        """
        Store statistics related to one run.

        :param statistics: The statistics to store.

        :return: None
        """

    @abstractmethod
    def add_signal(self, signal: Signal) -> None:
        """
        Store a signal.

        :param signal: THe signal to store.

        :return: None
        """


@versionadded(version="0.6.0")
@dataclass
class DataStoreDescription:
    """
    Entry point for an implementation of a data store backend.

    :param backend: The name of the backend the data store uses. This name has to be unique within the system.
    :type backend: str

    :param description: A textual description of the data store backend.
    :type description: str

    :param creation_fn: A factory function that can create the data store backend. Its parameters are the run_id and
        backend-specific key-value pairs.
    :type creation_fn: Callable[[str, Dict[str, str]], DataStore]
    """
    backend: str
    description: str
    creation_fn: Callable[[str, Dict[str, str]], DataStore]



from datetime import datetime
from typing import Any, Union, Callable

from cyst.api.environment.message import Timeout
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import Clock
from cyst.api.host.service import ActiveService

from cyst.platform.environment.message import TimeoutImpl

class SimulationClock(Clock):
    def __init__(self, messaging: EnvironmentMessaging):
        self._time = 0
        self._messaging = messaging

    def current_time(self) -> float:
        return self._time

    def real_time(self) -> datetime:
        raise NotImplementedError()

    def timeout(self, callback: Union[ActiveService, Callable[[Timeout], None]], delay: float, parameter: Any = None) -> None:
        timeout = TimeoutImpl(callback, self._time, delay, parameter)
        self._messaging.send_message(timeout, int(delay))
